"""A Python implementation of the EMuLSion framework (Epidemiologic
MUlti-Level SImulatiONs).

Exceptions raised by agents.
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


#  ______                    _   _
# |  ____|                  | | (_)
# | |__  __  _____ ___ _ __ | |_ _  ___  _ __  ___
# |  __| \ \/ / __/ _ \ '_ \| __| |/ _ \| '_ \/ __|
# | |____ >  < (_|  __/ |_) | |_| | (_) | | | \__ \
# |______/_/\_\___\___| .__/ \__|_|\___/|_| |_|___/
#                     | |
#                     |_|

class StateVarNotFoundException(Exception):
    """Exception raised when a semantic error occurs during model parsing.

    """
    def __init__(self, statevar, source):
        super().__init__()
        self.statevar = statevar
        self.source = source

    def __str__(self):
        return 'Statevar %s not found in object %s' % (self.statevar,
                                                       self.source)

class LevelException(Exception):
    """Exception raised when a semantic error occurs during model parsing.

    """
    def __init__(self, cause, level):
        super().__init__()
        self.level = level
        self.cause = cause

    def __str__(self):
        return 'Level %s %s' % (self.level, self.cause)


class InvalidCompartmentOperation(Exception):
    """Exception raised when a compartiment is asked for impossible
    operations, such as adding numbers to a list of units.

    """
    def __init__(self, source, operation, params):
        super().__init__(self)
        self.source = source
        self.operation = operation
        self.params = params

    def __str__(self):
        return "%s cannot execute '%s' with params: '%s'" %\
            (self.source, self.operation, self.params)



"""A Python implementation of the EMuLSion framework (Epidemiologic
MUlti-Level SImulatiONs).

Classes and functions for entities management.
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import numpy                     as np

from   emulsion.agent.core       import GroupAgent
from   emulsion.agent.exceptions import InvalidCompartmentOperation
from   emulsion.tools.misc       import POPULATION

class Compartment(GroupAgent):
    """An Compartment is a compartment which does not
    represent the underlying level but with aggregate information such
    as the total population ('individuals' are not represented).

    """
    def __init__(self, population=0, stochastic=True, **others):
        """Create an Compartment with an initial population."""
        super().__init__(**others)
        self.statevars.population = population
        self.stochastic = stochastic

    def __len__(self):
        return self.statevars.population

    def get_content(self):
        """Return the population of the current unit.

        """
        return ('population', self.statevars.population)

    def add(self, population):
        """Add the specified population to the current population of
        the compartment.

        """
        self.statevars.population += population

    def remove(self, population):
        """Remove the specified population from the current population
        of the compartment (the population is kept positive).

        """
        self.statevars.population = max(0, self.statevars.population - population)


    def _base_move(self, other_unit, population=0, **others):
        self.remove(population)
        other_unit.add(population)


    def move_to(self, other_unit, population, state_machine=None, **others):
        """Move the specified population from the current population
        of the compartment (the population is kept positive) to the
        other unit. If a state machine is provided, executes the
        corresponding actions when entering/exiting nodes and crossing
        edges if needed.

        """
        quantity = min(population, self.statevars.population)
        super().move_to(other_unit, population=quantity, state_machine=state_machine, **others)

    @property
    def population(self):
        return self.statevars.population

    def clone(self, **others):
        """Make a copy of the current compartment with the specified
        observable/value settings. The new content is empty.

        """
        new_comp = self.__class__.from_dict(self.statevars)
        new_comp.statevars.population = 0
        new_comp.model = self.model
        new_comp.stochastic = self.stochastic
        new_comp._host = self._host
        new_comp.statevars.update(**others)
        return new_comp

    def next_states(self, states, values, populations, actions, method=None):
        """Compute the population moving from the current compartment to each
        of the destination states, handling the values according the
        the specified method. Values can be handled either as absolute
        amounts ('amount' method), as proportions ('rate', in a
        deterministic approach) or as probabilities ('proba', in a
        stochastic approach). Actions are to be performed when
        changing state. The actual population affected by the
        transitions is stored in the first element of the
        `populations` parameter, as a dictionary: {'population':
        number, 'actions': actions}. Several edges can lead to the
        same state.

        Return a list of tuples:
          (state, {'population': qty, 'actions:' list of actions})
        """
        current_pop = populations[0][POPULATION]
        if method == 'amount':
            # length of values is expected to be the number of output edges
            # retrieve the amount of population exiting
            total_value = sum(values)
            if total_value > current_pop:
                # restart with proportions instead
                return self.next_states(states,
                                        tuple(v / total_value for v in values) + (0,),
                                        populations, actions, method=None)
            evolution = values
        else:
            if self.stochastic:
                # length of values is expected to be the number of
                # output edges + 1 (last value = 1 - sum(values[:-1])
                evolution = np.random.multinomial(current_pop, values)
            else:
                # length of values is expected to be the number of
                # output edges
                evolution = [(np.exp(rate*self.model.delta_t) - 1) * current_pop
                             for rate in values]
        result =  [(self._host.state_machine.states[state],
                    {'population': qty, 'actions': act})
                   for state, qty, act in zip(states[:-1], evolution, actions[:-1])
                   if qty > 0]
        return result



"""
.. module:: emulsion.agent.managers.functions

.. moduleauthor:: Sébastien Picault <sebastien.picault@inra.fr>

"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from   collections               import OrderedDict

from   sortedcontainers          import SortedDict

from   emulsion.agent.views      import StructuredView
from   emulsion.tools.misc       import count_population, rewrite_keys

from   emulsion.agent.managers.functions import group_and_split_populations

#   _____                       __  __
#  / ____|                     |  \/  |
# | |  __ _ __ ___  _   _ _ __ | \  / | __ _ _ __   __ _  __ _  ___ _ __
# | | |_ | '__/ _ \| | | | '_ \| |\/| |/ _` | '_ \ / _` |/ _` |/ _ \ '__|
# | |__| | | | (_) | |_| | |_) | |  | | (_| | | | | (_| | (_| |  __/ |
#  \_____|_|  \___/ \__,_| .__/|_|  |_|\__,_|_| |_|\__,_|\__, |\___|_|
#                        | |                              __/ |
#                        |_|                             |___/

class GroupManager(StructuredView):
    """An GroupManager is able to make its content
    evolve according to a specific state machine, the state of each
    subcompartment being stored in a specific state variable or
    attribute.

    """
    def __init__(self, state_machine=None, **others):
        """Create an GroupManager based on the
        specified state machine. The state of each subcompartment can
        be retrieved in the specified statevar name ('true' statevar
        or attribute)

        """
        ### WARNING: strange bug found sometimes when content={} not
        ### explicitly specified, another content (from another
        ### instance ???) may be used instead !!!!
        super().__init__(**others)
        self._content = SortedDict()
        self.state_machine = state_machine
        self.process_name = None
        self.init_counts()

    def init_counts(self, index=0):
        """Initialize the counts."""
        ## DEBUG
        # print('STEP', self.statevars.step, '\nInit counts in GM', self)
        self.counts = {}
        if self.state_machine is not None:
            self.counts = {state.name: [] if self.keep_history else 0
                           for state in self.state_machine.states}
            self.counts['step'] = [] if self.keep_history else 0
        else:
            super().init_counts()
        ## DEBUG
        # print(self.counts)

    def update_counts(self, index=0):
        """Update the number of atoms for each state of the state
        machine (TODO: for each value of the key[index] enum).

        """
        if self.state_machine is not None:
            total = {state.name: 0 for state in self.state_machine.states}
            ## DEBUG
            # print('\t', self.statevars.step, sep='')
            for (key, unit) in self._content.items():
                if key[index] is not None:
                    total[key[index].name] += unit.get_information('population')
                    # total[key[index].name] += unit.population
                    ## DEBUG
                    # print(key, unit.population, sep=' + ')
            # print()
            if self.keep_history:
                self.counts['step'].append(self.statevars.step)
                for state in self.state_machine.states:
                    self.counts[state.name].append(total[state.name])
            else:
                self.counts['step'] = self.statevars.step
                self.counts.update(total)
        else:
            super().update_counts()
        ## DEBUG
        # print('STEP', self.statevars.step, '\nUPDATE', self.counts)

    def apply_changes(self, transitions, productions):
        """Apply modifications to the compartments contained in the current
        StructuredView, according to `transitions` and
        `productions`. Dictionary `transitions` is keyed by a tuple of
        variables and associated with a list of dictionaries, either
        {'population': qty, 'actions': list} or {'agents': list,
        'actions': list}. List `productions` contains tuples (target,
        {'population': qty}, None) or (target, {'agents': list},
        prototype).

        """
        for source, evolutions in transitions.items():
            for target, population_or_agents in evolutions:
                target_comp = self.get_or_build(target, source=self[source])
                self._content[source].move_to(
                    target_comp,
                    state_machine=self.state_machine,
                    **population_or_agents)
        self.new_population = productions

    def evolve(self, machine=None):
        super().evolve(machine=machine)
        ## DEBUG
        # print(self.statevars.step, self.counts)
        self.evolve_states()
        ## DEBUG
        # print(self.statevars.step, self.counts)
        for key, comp in self._content.items():
            if comp.autoremove:
                agents_or_population = comp.get_content()
                if agents_or_population[0] == 'population':
                    agents_or_population = (agents_or_population[0],
                                            {self.process_name: {
                                                key: agents_or_population[1]
                                            }})
                self._host.remove(agents_or_population)
        self.update_counts()
        ## DEBUG
        # print(self.statevars.step, self.counts)

    def evolve_states(self, machine=None):
        """Ask each compartment to make its content evolve according
        to its current state and the specified state_machine.

        """
        self.new_population = None
        transitions = self._evolve_transitions(machine=machine)
        productions = self._evolve_productions(machine=machine)
        self.apply_changes(transitions, productions)

    def _evolve_transitions(self, machine=None):
        # init empty dictionary for all changes to perform
        future = OrderedDict()
        # iterate over all compartments
        for name, compart in self._content.items():
            future[name] = []
            # compute the current population of each source compartment
            current_pop = compart.get_information('population')
            # no action if current pop <= 0
            if current_pop <= 0:
                continue
            # compute all possible transitions from the current state
            current_state = compart.get_information(
                self.state_machine.machine_name)
            # execute actions on stay for current state
            compart.do_state_actions('on_stay', self.state_machine,
                                     current_state.name,
                                     **dict([compart.get_content()]))
            # get the possible transitions from the current state
            # i.e. a list of tuples (state, flux, value, cond_result,
            # actions) where:
            # - state is a possible state reachable from the current state
            # - flux is either 'rate' or 'proba' or 'amount' or 'amount-all-but'
            # - value is the corresponding rate or probability or amount
            # - cond_result is a tuple (either ('population', qty) or
            # ('agents', list)) describing who fulfills the condition to cross
            # the transition
            # - actions is the list of actions on cross
            transitions = compart.next_states_from(current_state.name,
                                                   self.state_machine)
            # print('TRANSITIONS = ', name, '->', transitions)
            # nothing to do if no transitions
            if not transitions:
                continue

            ### REWRITE TRANSITIONS TO HAVE DISJOINT SUB-POPULATIONS
            transitions_by_pop = group_and_split_populations(transitions)
            # print(transitions_by_pop)
            for ref_pop, properties in transitions_by_pop:
                # retrieve the list of states, the list of flux, the
                # list of values, the list of populations affected by
                # each possible transition
                states, flux, values, actions = zip(*properties)
                # print(name, '->\n\t', states, values, [ag._agid
                #                                        for u in populations
                #                                        for ag in u['agents']])
                # add the current state to the possible destination states...
                states = states + (current_state.name,)
                # ... with no action
                actions = actions + ([], )
                #
                values, method = self._compute_values_for_unique_population(
                    values, flux, ref_pop, compart.stochastic)
                change_list = compart.next_states(states,
                                                  values,
                                                  [ref_pop],
                                                  actions, method=method)
                future[name] += rewrite_keys(name, name.index(current_state),
                                              change_list)
        # print('FUTURE:', future)
        return future


    def _evolve_productions(self, machine=None):
        # init empty list for all changes to perform
        future = []
        # iterate over all compartments
        for name, compart in self._content.items():
            # compute the current population of each source compartment
            current_pop = max(compart.get_information('population'), 0)
            # no action if "fake" compartment
            if set(name) == {None}:
                continue
            # if current pop == 0, productions are still possible (e.g. amounts)
            # compute all possible transitions from the current state
            current_state = compart.get_information(self.state_machine.machine_name)
            # get the possible productions from the current state
            # i.e. a list of tuples (state, flux, value, cond_result,
            # prototype) where:
            # - state is a possible state producible from the current state
            # - flux is either 'rate' or 'proba' or 'amount' or 'amount-all-but'
            # - value is the corresponding rate or probability or amount
            # - cond_result is a tuple (either ('population', qty) or
            # ('agents', list)) describing who fulfills the condition to cross
            # the transition
            # - prototype is the prototype for creating new agents
            productions = compart.production_from(current_state.name, self.state_machine)
            # print('PRODUCTIONS = ', productions)
            # nothing to do if no transitions
            if not productions:
                continue
            ### HERE WE ASSUME THAT AN AGENT CAN PRODUCE SEVERAL OTHER
            ### AGENTS SIMULTANEOUSLY (OTHERWISE USE CONDITIONS)
            ### REWRITE TRANSITIONS TO HAVE DISJOINT SUB-POPULATIONS
            for target_state, flux, values, ref_pop, proto in productions:
                pop_size = count_population(ref_pop)
                amount = self._compute_production(values, flux, pop_size, compart.stochastic)
                if amount > 0:
                    future.append((target_state, amount, proto))
        # print('FUTURE:', future)
        return future



"""A Python implementation of the EMuLSion framework (Epidemiologic
MUlti-Level SImulatiONs).

Classes and functions for abstract agent management.

Part of this code was adapted from the PADAWAN framework (S. Picault,
Univ. Lille).
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import abc

from   sortedcontainers           import SortedSet


#  __  __      _                                 _
# |  \/  |    | |          /\                   | |
# | \  / | ___| |_ __ _   /  \   __ _  ___ _ __ | |_
# | |\/| |/ _ \ __/ _` | / /\ \ / _` |/ _ \ '_ \| __|
# | |  | |  __/ || (_| |/ ____ \ (_| |  __/ | | | |_
# |_|  |_|\___|\__\__,_/_/    \_\__, |\___|_| |_|\__|
#                                __/ |
#                               |___/

################################################################
# Metaclass for all agents
################################################################
class MetaAgent(abc.ABCMeta):
    """The Metaclass definition for all agents. When created, agents
    are stored in a class-specific dictionaries of agents. They are
    given an ID value (unique value within each class) and can be
    assigned to several agents families (by default, each agent is
    assigned to its own class).

    """
    @classmethod
    def __prepare__(mcs, name, bases, **kwds):
        families = SortedSet()
        families.add(name)
        attrs = {'agcount': 0,               # number of instances created
                 'agdict': {},               # dict of instances (ID -> agent)
                 'families': families}        # families where the class belongs
        # state NB: no need to keep information on passivity since it
        # depends on the families of the agent
        return attrs

    def __new__(mcs, name, bases, attrs, **_):
        attrs = dict(attrs)
        result = super(MetaAgent, mcs).__new__(mcs, name, bases, dict(attrs))
        result.members = tuple(attrs)
        return result




#!/usr/bin/env python3


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
from   pathlib    import Path
from   shutil     import copy

from   colorama   import Style
import emulsion

COMMAND = 'source $HOME/.emulsionrc/emulsion-completion.sh'
CPATH = 'export PYTHONPATH=$PYTHONPATH:.'

def main():
    # retrieve $HOME and path to EMULSION repository
    parts = Path(emulsion.__file__).parts[:-1] + ("scripts",)
    HOMEDIR = Path.home()
    EMULSION_SCRIPTS = Path(*parts)
    EMULSION_RC = Path(HOMEDIR, '.emulsionrc')
    # create .emulsionrc
    print("Creating $HOME/.emulsionrc with EMULSION init files")
    EMULSION_RC.mkdir(exist_ok=True)
    # copy emulsion-completion.sh
    copy(str(Path(EMULSION_SCRIPTS, 'emulsion-completion.sh')),
         str(EMULSION_RC))
    # change .bashrc
    print("Adding initialization instructions to $HOME/.bashrc")
    with open(Path(HOMEDIR, '.bashrc'), 'a') as f:
        f.write("\n# init EMULSION completion\n")
        f.write(COMMAND + "\n")
        f.write("\n# update PYTHONPATH to allow importing local modules\n")
        f.write(CPATH + "\n")

    print("Initialization finished.")
    print(Style.BRIGHT + """
    To allow completion for EMULSION in current shell, type:
    source $HOME/.bashrc
    """ + Style.RESET_ALL)

if __name__ == '__main__':
    main()



"""
.. module:: emulsion.agent.managers.functions

.. moduleauthor:: Sébastien Picault <sebastien.picault@inra.fr>

"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from   collections               import OrderedDict


#  ______                _   _
# |  ____|              | | (_)
# | |__ _   _ _ __   ___| |_ _  ___  _ __  ___
# |  __| | | | '_ \ / __| __| |/ _ \| '_ \/ __|
# | |  | |_| | | | | (__| |_| | (_) | | | \__ \
# |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/

def _group_by_populations(transitions):
    """`transitions` is a list of tuples (state, flux, value,
    cond_result, actions) where:
    - state is a possible state reachable from the current state flux
    - is either 'rate' or 'proba' or 'amount' or 'amount-all-but'
    - value is the corresponding rate or probability or amount
    - cond_result is a couple (either ('population', qty) or ('agents', list))
      describing who fulfills the condition to cross the transition
    - actions is the list of actions on cross
    """
    transitions_by_pop = OrderedDict()
    for state, flux, value, population, actions in transitions:
        desc, pop = population
        if desc == 'agents':
            pop = frozenset(pop)    # required to be hashable
        if (desc, pop) not in transitions_by_pop:
            transitions_by_pop[(desc, pop)] = []
        transitions_by_pop[(desc, pop)].append((state, flux, value, actions))
    # rewrite elements as a list
    return list(transitions_by_pop.items())

def _split_populations(list_of_items):
    """Transform a list of transitions into a dictionary based on the
    underlying populations.

    `list_of_items` is composed of pairs
    - the first element is a tuple ('agents', frozenset)
    - the second element is a list of tuples (state, flux, value,
      actions) associated with the population/agents

    Greedy algorithm: the first population of agents is compared with
    all others to check if any intersection exists. If so, both are
    replaced by their difference (if not empty) with the intersection,
    and the intersection is added to the list with a concatenation of
    the features attached to both populations.
    """
    # remove 'agents' keyword to facilitate set-like operations
    population_list = [(pop[1], attributes) for pop, attributes in list_of_items]
    i = 0
    while i < len(population_list) - 1:
        # select the first population in the list
        ref_pop, ref_attrib = population_list[i]
        j = i + 1
        intersects = False
        # check if any intersection with another population
        while j < len(population_list) and not intersects:
            # intersection found
            if population_list[j][0] & ref_pop:
                intersects = True
            else:
                j += 1
        if intersects:
            # retrieve features of the other population
            other_pop, other_attrib = population_list[j]
            # compute intersections and differences
            inter = ref_pop & other_pop
            new_ref =  ref_pop - inter
            new_other = other_pop - inter
            # if empty populations, remove them else update the populations
            if not new_other:
                population_list.pop(j)
            else:
                population_list[j] = (new_other, other_attrib)
            if not new_ref:
                population_list.pop(i)
            else:
                population_list[i] = (new_ref, ref_attrib)
            # add the intersection as a new population with a
            # concatenation of other properties
            population_list.append((inter, ref_attrib + other_attrib))
        else:
            # if no intersection was found, move to next item
            # (otherwise keep the same and iterate again on other
            # items)
            i += 1
    return [(('agents', list(agents)), attributes)
            for agents, attributes in population_list]


def group_and_split_populations(transitions):
    """Transform a list of transitions into a dictionary based on the
    underlying populations.

    `transitions` is a list of tuples (state, flux, value,
    cond_result, actions) where:
    - state is a possible state reachable from the current state flux
    - is either 'rate' or 'proba' or 'amount' or 'amount-all-but'
    - value is the corresponding rate or probability or amount
    - cond_result is a couple (either ('population', qty) or ('agents', list))
      describing who fulfills the condition to cross the transition
    - actions is the list of actions on cross

    The goal of this function is to restructure all those elements for
    disjoint sub-populations.

    Return a list of tuples:
      either (('population', qty), attributes)
      or     (('agents', list), attributes)
    where attributes is a list of tuples (state, flux, value, actions)

    """
    ### 1st step: gather elements
    # list_of_items is composed of pairs
    # - the first element is a tuple either ('population', qty) or
    #   ('agents', frozenset)
    # - the second element is a list of tuples (state, flux, value,
    #   actions) associated with the population/agents
    list_of_items = _group_by_populations(transitions)
    # if based on population (not on agents) then we're done (assuming
    # that all other values are based on populations)
    if list_of_items[0][0][0] == 'population':
        return list_of_items
    # otherwise we have true agents populations
    ### 2nd step: split sub-populations
    result =  _split_populations(list_of_items)
    #print(result)
    return result



""".. module:: emulsion.tools.state

This module defines:

- a class ``StateVarDict`` which is a special auto-referential
  dictionary to handle state variables
- a class ``EmulsionEnum`` which is a special kind of enumeration
  intended to handle states from state machines

"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from   enum                       import Enum
from   functools                  import total_ordering


class StateVarDict(dict):
    """A special dictionary aimed at handling the state variables of
    agents in EMULSION models. In addition to the classical dict
    key-based access, it provides an attribute-like access syntax.

    This class is used in EMULSION to store agent properties. Such
    properties include those defined automatically and used by the
    EMULSION engine, such as the values of the states for each state
    machine, the current time step, or "hidden" values such as the
    time spent in the current state for each state machine. They are
    also used to include user-defined attributes (e.g. age,
    weight...).

    When searching for a model ``statevar``, the engine tries first to
    find a classical instance variable (which can be mimicked by a
    Python ``@property``-decored function), then looks inside the
    agent's ``statevar`` attribute; finally, the search continues in
    the agent's host (if any).

    Example
    -------

    .. code-block:: python

        s = StateVarDict(age=10, sick=True)
        s['age'] += 1
        s.sick = False
        s.new_property = 'Wow !'

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # make self its own attributes dictionary (nice, hey ?)
        self.__dict__ = self


@total_ordering
class EmulsionEnum(Enum):
    """This class represents enumerations for states of state machines in
    EMULSION. They are endowed with some special features:

    #. They provide total ordering between items (based on ``__lt__``
       and ``__eq__`` methods).
    #. A comparison with ``None`` is provided (any state is always
       greater than ``None``).
    #. Other features will be developed soon.

    """

    def __lt__(self, other):
        """Return a less-than comparison for enumeration items. An enum item
        is always greater than ``None``and than any other item of
        another enumeration. Then, the ``value`` attribute of the item
        is used.

        Parameters
        ----------
        other: EmulsionEnum
            Another state from the same enumeration (or ``None``)

        Returns
        -------
        boolean:
            True if the current state is less than *other*
            (i.e. defined before *other* in the list of states), False
            otherwise (including the case where *other* is ``None``).

        """
        return False if other is None or other.__class__ != self.__class__\
          else self.value < other.value

    def __eq__(self, other):
        """Return a equality comparison for enumeration items. An enum item is
        always greater than ``None``and than any other item of another
        enumeration. Then, the ``value`` attribute of the item is
        used.

        Parameters
        ----------
        other: EmulsionEnum
            Another state from the same enumeration (or ``None``)

        Returns
        -------
        boolean:
            True if the current state is equal to *other* (having the
            same value), False otherwise (including the case where
            *other* is ``None``).

        """
        return False if other is None or other.__class__ != self.__class__\
          else self.value == other.value

    def __int__(self):
        """Return the *int* value mapped to this item."""
        return self.value

    def __repr__(self):
        """Return a string representation of the instances of the enumeration,
        hiding associated numerical value.

        """
        return f'<{self.__class__.__name__}.{self.name}>'

    def __hash__(self):
        """Use the value as hashcode."""
        return self.value



"""
.. module:: emulsion.agent.managers.functions

.. moduleauthor:: Sébastien Picault <sebastien.picault@inra.fr>

"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import itertools                 as it
from   collections               import OrderedDict

import numpy                     as np
import pandas                    as pd

from   emulsion.agent.comparts   import Compartment

from   emulsion.agent.managers.group_manager  import  GroupManager
from   emulsion.agent.managers.abstract_process_manager  import  AbstractProcessManager
from   emulsion.agent.exceptions import StateVarNotFoundException

#   _____                                 _   _____
#  / ____|                               | | |  __ \
# | |     ___  _ __ ___  _ __   __ _ _ __| |_| |__) | __ ___   ___ ___  ___ ___
# | |    / _ \| '_ ` _ \| '_ \ / _` | '__| __|  ___/ '__/ _ \ / __/ _ \/ __/ __|
# | |___| (_) | | | | | | |_) | (_| | |  | |_| |   | | | (_) | (_|  __/\__ \__ \
#  \_____\___/|_| |_| |_| .__/ \__,_|_|   \__|_|   |_|  \___/ \___\___||___/___/
#                       | |
#                       |_|
#  __  __
# |  \/  |
# | \  / | __ _ _ __   __ _  __ _  ___ _ __
# | |\/| |/ _` | '_ \ / _` |/ _` |/ _ \ '__|
# | |  | | (_| | | | | (_| | (_| |  __/ |
# |_|  |_|\__,_|_| |_|\__,_|\__, |\___|_|
#                            __/ |
#                           |___/

class CompartProcessManager(AbstractProcessManager):
    """A CompartProcessManager is aimed handling several independent
    StructuredViews at the same time, for managing true compartments.
    It can automatically allocate compartments for state machines
    associated with a specific state variable or attribute.

    """
    def add_compart_process(self,
                            process_name,
                            key_variables,
                            compart_manager=(GroupManager, {}),
                            machine_name=None,
                            compart_class=(Compartment, {})):
        super().add_compart_process(process_name, key_variables,
                                    compart_manager=compart_manager,
                                    machine_name=machine_name,
                                    compart_class=compart_class)

    def add_host(self, host):
        """Add the specified host to the current Multiprocessmanager, associated
        with the specified key.

        """
        if self._host is None:
            self._host = OrderedDict()
        self._host[host.keys] = host
        self.simulation = host.simulation

    def apply_initial_conditions(self):
        """Initialize level with initial conditions specified in the model.

        As this agent is aimed at managing aggregated populations in
        the sub-levels, only 'populations' are taken into account to
        initialize the sub-levels.

        """
        # print(self.keys)
        if self.level in self.model.initial_conditions:
            conds = self.model.initial_conditions[self.level]
            total = self.get_model_value(conds['total']) if 'total' in conds else 0
            population_to_add = {}
            # print(conds)
            for proc, compart in self._content.items():
                if proc in self.no_compart:
                    continue
                ## assuming here that in compartment-based models all
                ## keys are based on enumerations

                ## build the combinations of all possible compartment
                ## keys (excluding autoremove states)
                combinations = list(it.product(*[[state
                                                  for state in self.model.state_machines[machine_name].states
                                                  if not state.autoremove]
                                                 for machine_name in compart.keys]))
                # population_to_add[proc] = { key: 0 for key in combinations }
                population_to_add[proc] = {}
                subtotal = 0
                # compute initial values for each group
                for state_names, qty in conds['populations']:
                    key = tuple(self.get_model_value(state_name) for state_name in state_names)
                    # if key in population_to_add[proc]:
                    if key in combinations:
                        amount = self.get_model_value(qty)
                        population_to_add[proc][key] = amount
                        subtotal += amount

                total = max(total, subtotal)
            self.complement_population(population_to_add, total=total)
            # print(population_to_add)
            self.add_population(population_to_add, init=True)
            for proc, compart in self._content.items():
                if proc not in self.no_compart:
                    compart.update_counts()

    def add_new_population(self, process_name, population):
        to_add = {}
        for target_state, amount, _ in population:
            key = (self.get_model_value(target_state),)
            if key not in to_add and amount > 0:
                to_add[key] = amount
            else:
                to_add[key] += amount
        self.add_population({process_name: to_add})


    def complement_population(self, population_to_change, total=0, remove=False):
        """Modifiy the population spec in a consistent way. Check is all
        populations in each process sum to the same total. If not, distribute
        the difference between total and sum at random between all available
        groups (FAIL if no available groups).

        """
        ## complement keys
        for proc, compart in self._content.items():
            if proc in self.no_compart:
                continue
            ## assuming here that in compartment-based models all
            ## keys are based on enumerations

            ## build the combinations of all possible compartment
            ## keys (excluding autoremove states)
            if remove:
                combinations = list(
                    it.product(*[[state
                                  for state in self.model.state_machines[machine_name].states
                                  if not state.autoremove]
                                 for machine_name in compart.keys]))
            else:
                combinations = list(
                    it.product(*[self.model.state_machines[machine_name].states.available
                                 for machine_name in compart.keys]))
            if proc not in population_to_change:
                population_to_change[proc] = {}
            population_to_change[proc].update({ key: 0
                                                for key in combinations
                                                if key not in population_to_change[proc]})
        ## compute total per process
        total_values = {}
        for proc in self._content:
            if proc in self.no_compart:
                continue
            total_values[proc] = sum(population_to_change[proc].values())\
                                 if proc in population_to_change else 0
        total = max(total, *total_values.values())

        # any process with less than total ?
        if any(value < total for value in total_values.values()):
            for proc, compart in self._content.items():
                if proc in self.no_compart:
                    continue
                subtotal = sum(population_to_change[proc].values())
                if subtotal < total:
                    difference = total - subtotal
                    available = [key for key in population_to_change[proc].keys()
                                 if population_to_change[proc][key] == 0]
                    if remove:
                        available = [key for key in available
                                     if key in compart._content and\
                                     compart[key].statevars.population > 0]
                    if not available:
                        print("WARNING !!! initial populations for process {} is"
                              " less than total population {} and there is no"
                              " empty compartment to redistribute the"
                              " difference".format(proc, subtotal))
                    else:
                        nb = len(available)
                        if remove:
                            pops = [compart[key].statevars.population for key in available]
                            probas = [pop / sum(pops) for pop in pops]
                        else:
                            probas = [1 / nb] * nb

                        amounts = np.random.multinomial(difference, probas)\
                                  if self.stochastic\
                                  else [difference * p for p in probas]

                        for key, amount in zip(available, amounts):
                            population_to_change[proc][key] = amount



    def add_population(self, population_spec, init=False):
        """Add the specified population specification to the current
        CompartProcessManager. `population_spec` is a dictionary with
        process names as keys, each one associated with a dictionary
        (tuple of statevars) -> population. If `init` is True, the compartment
        managers counts the initial value of the populations in each
        compartment.

        """
        self.complement_population(population_spec)
        # add populations to appropriate compartments for each process
        added = {}
        for process, spec in population_spec.items():
            # print(process, spec)
            if process not in self.no_compart:
                # retrieve the group manager associated to the process
                manager = self[process]
                # retrieve the machine name (always here in compartment models)
                machine_name = manager.state_machine.machine_name
                # locate the index of the statevar holding the state
                # of the state machine
                index = manager.keys.index(machine_name)\
                        if machine_name in manager.keys else None
                default_key = tuple(None for _ in manager.keys)
                added[process] = 0
                for key, qty in spec.items():
                    if key not in manager._content:
                        new_comp = manager[default_key].clone(population=qty)
                        if index is not None:
                            new_comp.statevars[machine_name] = key[index]
                        new_comp.keys = key
                        manager._content[key] = new_comp
                    else:
                        manager[key].add(qty)
                    added[process] += qty
        if init:
            manager.update_counts()
        # print(list(added.values()))
        nb = set(added.values())
        assert(len(nb) <= 1)
        if nb:
            self.statevars.population += nb.pop()

    def remove(self, agents_or_population):
        assert(agents_or_population[0] == 'population')
        self.remove_population(agents_or_population[1])

    def remove_population(self, population_spec):
        """Remove the specified population spec from the current
        CompartProcessManager.

        """
        self.complement_population(population_spec, remove=True)
        # print('>'*10)
        # print(population_spec)
        # print('<'*10)
        removed = {}
        for process, spec in population_spec.items():
            if process not in self.no_compart:
                # print(process, spec)
                manager = self[process]
                removed[process] = 0
                for key, qty in spec.items():
                    if key in manager._content:
                        manager[key].remove(qty)
                        removed[process] += qty
        nb = set(removed.values())
        assert(len(nb) <= 1)
        if nb:
            self.statevars.population -= nb.pop()

    def remove_randomly(self, proba=0, amount=None, process=None):
        """Remove random amounts of populations from this ProcessManager. If
        `amount` is not None, a multinomial sampling is performed for
        each process. Otherwise: `proba` can be either a probability
        or a dictionary. In that case, the `process` parameter
        indicates the name of the process grouping which drives the
        probabilities, and the keys must be those of the
        grouping. Selected quantities are removed and returned by the
        method.

        """
        to_remove = {}
        total = None
        if amount is not None:
            for name, proc in self._content.items():
                if name not in self.no_compart:
                    keys, probs = zip(*[(key, comp.population)
                                        for key, comp in proc._content.items()])
                    s = sum(probs)
                    probas = [p / s for p in probs]
                    amounts = np.random.multinomial(amount, probas)
                    to_remove[name] = dict(zip(keys, amounts))
                    # print(to_remove)
            self.remove_population(to_remove)
            return to_remove

        if process is not None:
            to_remove[process] = {}
            total = 0
            # print(self._content, process, name, self.no_compart)
            for key, comp in self[process].items():
                pop = comp.statevars.population
                if key in proba:
                    n = np.random.binomial(pop, proba[key])
                    to_remove[process][key] = n
                    total += n
        for name, proc in self._content.items():
            if name in self.no_compart or name == process:
                continue
            if total is None:
                to_remove[name] = {}
                total = 0
                for key, comp in proc._content.items():
                    pop = comp.statevars.population
                    n = np.random.binomial(pop, proba)
                    to_remove[name][key] = n
                    total += n
                continue
            keys, pops = zip(*[(key, comp.statevars.population)
                               for key, comp in self[name].items()])
            total_pop = sum(pops)
            probas = [ n / total_pop for n in pops]
            qties = np.random.multinomial(total, probas)
            to_remove[name] = dict(zip(keys, qties))
        # print(to_remove)
        self.remove_population(to_remove)
        return to_remove

    @property
    def counts(self):
        """Return a pandas DataFrame containing counts of each process if existing.
        TODO: column steps need to be with one of process

        """
        res = {}
        for comp in self:
            try:
                res.update(comp.counts)
            except AttributeError:
                pass
            except Exception as exc:
                raise exc
        if not self.keep_history:
            res.update({
                'level': self.level,
                'agent_id': self.agid,
                # 'population': self.population}
            })
            if self.level in self.model.outputs and\
               'extra_vars' in self.model.outputs[self.level]:
                for name in self.model.outputs[self.level]['extra_vars']:
                    if name in self.model.parameters:
                        res[name] = self.get_model_value(name)
                    else:
                        try:
                            value = self.get_information(name)
                        except StateVarNotFoundException:
                            value = np.nan
                        res[name] = value
        return pd.DataFrame(res, index=[0])



"""
.. module:: emulsion.agent.core.groups

.. moduleauthor:: Sébastien Picault <sebastien.picault@inra.fr>

Part of this code was adapted from the PADAWAN framework (S. Picault,
Univ. Lille).
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc

from   emulsion.agent.core.emulsion_agent  import EmulsionAgent

#   _____                                               _
#  / ____|                        /\                   | |
# | |  __ _ __ ___  _   _ _ __   /  \   __ _  ___ _ __ | |_
# | | |_ | '__/ _ \| | | | '_ \ / /\ \ / _` |/ _ \ '_ \| __|
# | |__| | | | (_) | |_| | |_) / ____ \ (_| |  __/ | | | |_
#  \_____|_|  \___/ \__,_| .__/_/    \_\__, |\___|_| |_|\__|
#                        | |            __/ |
#                        |_|           |___/

class GroupAgent(EmulsionAgent):
    """A GroupAgent is aimed at representing a group of agents. The
    underlying level may be either explicitly represented (using a
    ViewCompartment) or aggregated (using an Compartment). Each
    compartment is associated with keys, i.e. a tuple of state
    variables (possibly empty) which play a crucial role in this
    compartment.

    """
    def __init__(self, keys=(), **others):
        super().__init__(**others)
        self.keys = keys
        self.autoremove = False


    @abc.abstractmethod
    def add(self, population):
        """Add the specified population to the current compartment."""
        pass

    @abc.abstractmethod
    def remove(self, population):
        """Remove the specified population from the current
        compartment.

        """
        pass

    @abc.abstractmethod
    def _base_move(self, other_unit, **others):
        pass

    def _before_move(self, state_machine, old_state, new_state, **others):
        # execute actions when exiting current state (if any)
        self.do_state_actions('on_exit',
                              state_machine,
                              old_state,
                              **others)
        # execute actions when crossing edge (if any)
        self.do_edge_actions(**others)
        # update states of atom units
        if 'agents' in others:
            for unit in others['agents']:
                unit.change_state(state_machine.machine_name,
                                  state_machine.states[new_state])
                # unit.statevars[state_machine.machine_name] =\
                #   state_machine.states[new_state]

    def _after_move(self, state_machine, new_state, **others):
        # execute actions when entering new state (if any)
        self.do_state_actions('on_enter',
                              state_machine,
                              new_state,
                              **others)



    def move_to(self, other_unit, state_machine=None, **others):
        """Move the specified population from the current population
        of the compartment to the other unit. If a state machine is
        provided, executes the corresponding actions when
        entering/exiting nodes and crossing edges if needed.

        """
        if state_machine:
            #old_state = self.get_information(state_machine.machine_name).name
            old_state = self.statevars[state_machine.machine_name]
            new_state = other_unit.statevars[state_machine.machine_name].name
            if old_state is not None:
                self._before_move(state_machine, old_state.name,
                                  new_state, **others)
        # move population from current compartment to other unit
        self._base_move(other_unit, **others)
        if state_machine:
            other_unit._after_move(state_machine, new_state, **others)



#                                           _   _
#     /\                                   | | (_)
#    /  \   __ _  __ _ _ __ ___  __ _  __ _| |_ _  ___  _ __
#   / /\ \ / _` |/ _` | '__/ _ \/ _` |/ _` | __| |/ _ \| '_ \
#  / ____ \ (_| | (_| | | |  __/ (_| | (_| | |_| | (_) | | | |
# /_/    \_\__, |\__, |_|  \___|\__, |\__,_|\__|_|\___/|_| |_|
#           __/ | __/ |          __/ |
#          |___/ |___/          |___/


class Aggregation(GroupAgent):
    """An Aggregation is aimed at grouping agents from the underlying
    level. Thus, aggregate information such as the total number of
    agents in the compartment (population) is calculated from the
    actual content of the compartment. The evolution of a
    ViewCompartment, by default, consists in making the units
    contained in the compartment evolve themselves, unless
    `recursive=False` is specified during instantiation.

    """
    def __init__(self, recursive=True, **others):
        super().__init__(**others)
        self._content = None
        self.recursive = recursive

    @property
    def population(self):
        """Return the total population of the compartment. It is
        calculated either using a true 'population' statevar if any,
        or as the sum of the population of each unit contained in the
        compartment.

        TAG: USER
        """
        ## DEBUG
        # print('\t\tstatevars.population:', self.statevars.population if 'population' in self.statevars else 'None', [f'{unit.agid}->{unit._host}' for unit in self])
        return self.statevars.population if 'population' in self.statevars\
            else sum([unit.get_information('population')
                      for unit in self])

    def add(self, population):
        """Add the specified population to the current compartment."""
        for unit in population:
            unit.add_host(self)

    def remove(self, population):
        """Remove the specified population from the current
        compartment.

        """
#        print([u.agid for u in population])
        for unit in population:
            unit.remove_host(self, keys=self.keys)


    def _base_move(self, other_unit, agents=[], **others):
        self.remove(agents)
        other_unit.add(agents)


    @abc.abstractmethod
    def __iter__(self):
        pass

    # TODO: shuffle content first
    def evolve(self, machine=None):
        """Ask each unit in the current compartment to make its content evolve
        according to its own capabilities. A specific state machine
        can be specified if needed.

        """
        super().evolve(machine=machine)
        if self.recursive:
            for unit in self:
                unit.evolve(machine=machine)



"""
.. module:: emulsion.agent.managers.functions

.. moduleauthor:: Sébastien Picault <sebastien.picault@inra.fr>

"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc
import itertools                 as it

from   sortedcontainers          import SortedSet

import numpy                     as np
import pandas                    as pd

from   emulsion.agent.views      import StructuredView, AdaptiveView
from   emulsion.agent.process    import MethodProcess, StateMachineProcess
from   emulsion.agent.exceptions import StateVarNotFoundException
from   emulsion.tools.misc       import load_class, add_new_property,\
    create_population_getter, add_all_test_properties, create_aggregator,\
    create_group_aggregator


from   emulsion.agent.managers.group_manager  import  GroupManager


#           _         _                  _
#     /\   | |       | |                | |
#    /  \  | |__  ___| |_ _ __ __ _  ___| |_
#   / /\ \ | '_ \/ __| __| '__/ _` |/ __| __|
#  / ____ \| |_) \__ \ |_| | | (_| | (__| |_
# /_/    \_\_.__/|___/\__|_|  \__,_|\___|\__|
#  _____                             __  __
# |  __ \                           |  \/  |
# | |__) | __ ___   ___ ___  ___ ___| \  / | __ _ _ __   __ _  __ _  ___ _ __
# |  ___/ '__/ _ \ / __/ _ \/ __/ __| |\/| |/ _` | '_ \ / _` |/ _` |/ _ \ '__|
# | |   | | | (_) | (_|  __/\__ \__ \ |  | | (_| | | | | (_| | (_| |  __/ |
# |_|   |_|  \___/ \___\___||___/___/_|  |_|\__,_|_| |_|\__,_|\__, |\___|_|
#                                                              __/ |
#                                                             |___/

class AbstractProcessManager(StructuredView):
    """An AbstractProcessManager is aimed handling several independent
    StructuredViews at the same time, to represent several
    concerns. It can automatically build compartments for state
    machines associated with a specific state variable or attribute.

    """
    def __init__(self, model=None, master=None, level=None, stochastic=True,
                 keep_history=False, prototype=None, custom_prototype=None,
                 execute_actions=False, **others):
        super().__init__(recursive=False, **others)
        self.statevars.population = 0
        self.stochastic = stochastic
        self.model = model
        self.level = level
        self.keep_history = keep_history
        if master is not None:
            self._content['MASTER'] = master
            self.no_compart = SortedSet(['MASTER'])
        else:
            self.no_compart = SortedSet()
        # dynamically add property 'total_level' which returns the
        # total population in this level, and aggegate variables if any
        if self.level is not None:
            add_new_property(self, 'total_{}'.format(self.level),
                             lambda self: self.get_information('population'))
            if self.level in self.model.aggregate_vars:
                for (varname, sourcevar, operator) in self.model.aggregate_vars[self.level]:
                    add_new_property(self, varname, create_aggregator(sourcevar, operator))
        # machine names that potentially require state initialization
        self.init_machines = SortedSet()
        self.init_processes()
        ## apply prototype to self before creating sublevels
        ## done with super() method to control when it happens
        super().apply_initial_prototype(name=prototype, prototype=custom_prototype, execute_actions=execute_actions)
        self.apply_initial_conditions()
        self.initialize_level()

    def load_state_from_file(self, simu_id, filename):
        #TODO
        pass

    def save_state_to_file(self, simu_id, filename):
        #TODO
        pass

    def apply_initial_prototype(self, name=None, prototype=None, execute_actions=False):
        """This method inherited from `AbstractAgent` and called by `new_atom`
        is intentionnaly doing nothing in the current class to ensure
        that initial prototype of a `MultiProcessManager` is applied
        before creating the sublevels.

        """
        pass

    @abc.abstractmethod
    def apply_initial_conditions(self):
        """Apply initial conditions (if any) as defined in the model."""
        pass

    def initialize_level(self, **others):
        """User-defined operations when creating an instance of this
        level. These operations are performed *after* the application of
        initial conditions possibly defined in the corresponding model
        section."""
        pass

    def finalize_level(self, simulation=None, **others):
        """User-defined operations at the end of simulation for an instance of
        this level.

        """
        pass

    # @abc.abstractmethod
    def add_new_population(self, process_name, population):
        pass

    def init_processes(self):
        """Init the processes that the ProcessManager will undergo during each
        time step, in order. Processes may be either 'method'
        processes (based on the execution of the specified method
        name), or 'group-based' processes (defined by the evolution of
        a grouping (aggregation or compartment), possibly endowed with
        a state machine), or even a 'state-machine-driven' process,
        based on the direct execution of a state machine within the
        ProcessManager.

        """
        if self.level in self.model.processes:
            for process in self.model.processes[self.level]:
                if self.level not in self.model.compartments or\
                   process not in self.model.compartments[self.level]:
                    if process in self.model.state_machines:
                        self.add_statemachine_process(process)
                    else:
                        self.add_method_process(process)
                else:
                    compart_properties = dict(
                        self.model.compartments[self.level][process])
                    for keyword in ['compart_manager', 'compart_class']:
                        if keyword in compart_properties:
                            class_desc = compart_properties[keyword]
                            # print(keyword, class_desc)
                            compart_properties[keyword] = load_class(**class_desc)
                    self.add_compart_process(process, **compart_properties)

    def evolve(self, **others):
        """Make the ProcessManager evolve, i.e. all the registered processes
        in order, starting with the evolution of the sublevels, and
        followed by the evolution inherited from superclasses.

        """
#        self['MASTER'].evolve()
        for process in self:
            process.evolve()
        for name, process in self._content.items():
            if name not in self.no_compart:
                self.add_new_population(name, process.new_population)
                process.update_counts()
        super().evolve(**others)

    def add_method_process(self, process_name, method=None):
        """Add a process based on a method name."""
        # print('process:', process_name)
        if method is None:
            method = getattr(self, process_name)
        self._content[process_name] = MethodProcess(process_name, method)
        self.no_compart.add(process_name)

    def add_statemachine_process(self, process_name):
        """Add a process based on the direct execution of a state machine."""
        self._content[process_name] = StateMachineProcess(
            process_name, self, self.model.state_machines[process_name]
        )
        self.no_compart.add(process_name)

    def add_compart_process(self,
                            process_name,
                            key_variables,
                            compart_manager=(GroupManager, {}),
                            machine_name=None,
                            allowed_values=None,
                            compart_class=(AdaptiveView, {})):
        """Add a process aimed at managing a 'Compartment Manager', i.e. an
        object aimed at managing a collection of compartments. This
        compartment manager is automatically initialized from the
        `compart_manager` class (which should be a subclass of
        StructuredView or GroupManager). The compartment manager may
        be associated with a specific state machine, and MUST BE
        identified by a tuple of state variables names. Additionally,
        since a first compartment is also instantiated, a specific
        class to do so can be also specified.

        """
        args = {'keys': tuple(key_variables), 'host': self,
                'keep_history': self.keep_history}
        compart_manager_cl, manager_options = compart_manager
        compart_class_cl, compart_options = compart_class
        if machine_name:
            machine = self.model.state_machines[machine_name]
            args['state_machine'] = machine
            # print(machine_name)
            if machine_name in self.model.init_actions:
                self.init_machines.add(machine_name)
            # dynamically add properties for accessing counts of each state
            for state in machine.states:
                if not state.autoremove:
                    self.create_count_properties_for_state(process_name,
                                                           state.name,
                                                           create_population_getter,
                                                           create_group_aggregator)
        if allowed_values:
            args['allowed_values'] = allowed_values
        args.update(manager_options)
        dict_comp = compart_manager_cl(**args)
        # update the model of the compartment manager
        dict_comp.model = self.model
        dict_comp.process_name = process_name
        init_key = tuple(None for _ in key_variables)
        dict_comp._content[init_key] = compart_class_cl(
            recursive=False,
            stochastic=self.stochastic,
            observables=key_variables,
            keys=init_key,
            values=init_key,
            host=dict_comp, **compart_options)
        # update the model of views
        dict_comp._content[init_key].model = self.model
        self._content[process_name] = dict_comp
        # dynamically add properties for accessing sub-groups when
        # groupings are based on several states from statemachines
        if len(key_variables) > 1 and\
           all(key in self.model.state_machines for key in key_variables):
            self.create_properties_for_groups(process_name, key_variables)
        # dynamically add properties for testing states in compart_class instances
        add_all_test_properties(dict_comp._content[init_key])

    def create_count_properties_for_state(self, process_or_machine_name, state_name,
                                          count_function, aggregation_function):
        """Dynamically add properties of the form ``total_S`` where S can be
        any state of the state machine. Counts are expected to be
        computed by the grouping associated with the specified
        process. The access to counts is defined by
        *count_function*. If aggregated variables such as ``aggvar``
        are defined, corresponding properties of the form
        ``aggvar_S``are also defined with the specified
        *aggregation_function*.

        """
        add_new_property(self, 'total_{}'.format(state_name),
                         count_function(process_or_machine_name, state_name))
        if self.level in self.model.aggregate_vars:
            for (varname, sourcevar, operator) in self.model.aggregate_vars[self.level]:
                add_new_property(self, '{}_{}'.format(varname, state_name),
                                 aggregation_function(sourcevar, operator,
                                                      process_or_machine_name, state_name))

    def create_properties_for_groups(self, process_name, key_variables):
        """Dynamically add properties of the form `total_S_T` where S, T are a
        valid key in grouping associated with the specified
        process_name.

        """
        # print(key_variables)
        combinations = list(it.product(*[[state.name
                                          for state in self.model.state_machines[machine_name].states
                                          if not state.autoremove]
                                         for machine_name in key_variables]))
        for group in combinations:
            add_new_property(self, 'total_{}'.format('_'.join(group)),
                             create_population_getter(process_name, group))
            if self.level in self.model.aggregate_vars:
                for (varname, sourcevar, operator) in self.model.aggregate_vars[self.level]:
                    add_new_property(self, '{}_{}'.format(varname, '_'.join(group)),
                                     create_group_aggregator(sourcevar, operator,
                                                             process_name, group))


    def get_group_population(self, process_name, group_name):
        """Return the size of the subgroup associated with the specified group
        name (state names).

        """
        complete_key = set(self.get_model_value(state_name)
                           for state_name in group_name)
        groups = self[process_name]
        value = 0
        for key, compart in groups._content.items():
            if complete_key <= set(key):
                value += compart.population
        return value

    @property
    def counts(self):
        """Return a pandas DataFrame containing counts of each process if existing.
        TODO: column steps need to be with one of process

        """
        res = {}
        for comp in self:
            try:
                # TODO: BUG with StructuredViewWithCounts, step=0 always !
                if comp.__class__.__name__ != 'StructuredViewWithCounts':
                    comp.update_counts()
                    res.update(comp.counts)
            except AttributeError:
                pass
            except Exception as exc:
                raise exc
        if not self.keep_history:
            res.update({
                'level': self.level,
                'agent_id': self.agid,
                # 'population': self.population
            })
            if self.level in self.model.outputs and\
               'extra_vars' in self.model.outputs[self.level]:
                for name in self.model.outputs[self.level]['extra_vars']:
                    if name in self.model.parameters:
                        res[name] = self.get_model_value(name)
                    else:
                        try:
                            value = self.get_information(name)
                        except StateVarNotFoundException:
                            value = np.nan
                        res[name] = value
        df = pd.DataFrame(res, index=[0])
        ## DEBUG
        # print(df)
        return df

    @abc.abstractmethod
    def remove_randomly(self, proba=0):
        """Remove randomly chosen atoms or population from this
        ProcessManager.

        """
        pass

    @abc.abstractmethod
    def remove(self, agents_or_population):
        pass

    @property
    def population(self):
        return self.statevars.population



"""
.. module:: emulsion.model.state_machines

.. moduleauthor:: Sébastien Picault <sebastien.picault@inra.fr>

"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from   functools               import partial

import numpy                   as     np
from   sympy                   import sympify

from   sortedcontainers        import SortedSet

import emulsion.tools.graph    as     enx
from   emulsion.agent.action   import AbstractAction
from   emulsion.tools.state    import StateVarDict, EmulsionEnum

from   emulsion.model.functions import ACTION_SYMBOL, WHEN_SYMBOL, ESCAPE_SYMBOL,\
    COND_SYMBOL, CROSS_SYMBOL, EDGE_KEYWORDS, CLOCK_SYMBOL,\
    make_when_condition, make_duration_condition, make_duration_init_action

from   emulsion.model.exceptions     import SemanticException


#   _____ _        _       __  __            _     _
#  / ____| |      | |     |  \/  |          | |   (_)
# | (___ | |_ __ _| |_ ___| \  / | __ _  ___| |__  _ _ __   ___
#  \___ \| __/ _` | __/ _ \ |\/| |/ _` |/ __| '_ \| | '_ \ / _ \
#  ____) | || (_| | ||  __/ |  | | (_| | (__| | | | | | | |  __/
# |_____/ \__\__,_|\__\___|_|  |_|\__,_|\___|_| |_|_|_| |_|\___|



class StateMachine(object):
    """Class in charge of the description of biological or economical
    processes, modeled as Finite State Machines. The formalism
    implemented here is based on UML state machine diagrams, with
    adaptations to biology.

    """
    def __init__(self, machine_name, description, model):
        """Build a State Machine within the specified model, based on
        the specified description (dictionary).

        """
        self.model = model
        self.machine_name = machine_name
        self.parse(description)

    def _reset_all(self):
        self._statedesc = {}
        self._description = {}
        self.states = None
        self.graph = enx.MultiDiGraph()
        self.stateprops = StateVarDict()
        self.state_actions = {}
#        self.edge_actions = {}

    def parse(self, description):
        """Build the State Machine from the specified dictionary
        (expected to come from a YAML configuration file).

        """
        self._reset_all()
        # keep an exhaustive description
        self._description = description
        # build the enumeration of the states
        self.build_states()
        # build the graph based on the states and the transitions between them
        self.build_graph()
        # build actions associated with the state machine (states or edges)
        self.build_actions()

    def get_property(self, state_name, property_name):
        """Return the property associated to the specified state."""
        if state_name not in self.stateprops or\
           property_name not in self.stateprops[state_name]:
            return self.graph.node[state_name][property_name]\
                if property_name in self.graph.node[state_name]\
                   else None
        return self.stateprops[state_name][property_name]

    def build_states(self):
        """Parse the description of the state machine and extract the existing
        states. States are described as list items, endowed with
        key-value properties. Only one state per list item is allowed
        (to ensure that states are always stored in the same order in
        all executions).

        """
        states = []
        default_state = None
        # retrieve information for each state
        for statedict in self._description['states']:
            for name, value in statedict.items():
                states.append(name)
                # provide a default fillcolor
                if 'fillcolor' not in value:
                    value['fillcolor'] = 'lightgray'
                # if properties are provided, add the corresponding
                # expression to the model
                if 'properties' not in value:
                    value['properties'] = {}
                # store special property: "autoremove: yes"
                value['properties']['autoremove'] = value['autoremove']\
                                                    if 'autoremove' in value else False
                # store special property: "default: yes"
                # if several states are marked "default", take the first one
                value['properties']['default'] = False
                if ('default' in value) and (value['default']) and (default_state is None):
                    value['properties']['default'] = True
                    default_state = name
                self.stateprops[name] = {k: self.model.add_expression(v)
                                         for k, v in value['properties'].items()}
                # store special properties: "next: aState" and
                # "previous: aState" which define
                # successor/predecessor when using "next_state" and
                # "previous_state" in prototype definition (otherwise,
                # next and previous are defined as non-autoremove
                # states after and before in the list)
                if 'next' in value:
                    value['properties']['next'] = value['next']
                    self.stateprops[name]['next'] = value['next']
                if 'previous' in value:
                    value['properties']['previous'] = value['previous']
                    self.stateprops[name]['previous'] = value['previous']
                # store other information
                self._statedesc[name] = value
                # and retrieve available actions if any
                for keyword in ['on_enter', 'on_stay', 'on_exit']:
                    if keyword in value:
                        self._add_state_actions(name, keyword, value[keyword])
        # build the enumeration of the states
        self.states = EmulsionEnum(self.machine_name.capitalize(),
                                   states, module=__name__)
        # link the states to their state machine
        self.states.state_machine = self
        # define the default value for "autoremove" at the enumeration level
        self.states.autoremove = False

        for state in self.states:
            # check that state names are unique
            if state.name in self.model.states:
                other_machine = self.model.states[state.name].__class__.__name__
                raise SemanticException(
                    'Conflict: State %s found in statemachines %s and %s' %
                    (state.name, other_machine, state.__class__.__name__))
            # check that state names are not parameters
            if state.name in self.model.parameters:
                raise SemanticException(
                    'Conflict: State %s of statemachines %s found in parameters'
                    % (state.name, state.__class__.__name__))
            # associate the state with the state name in the model
            self.model.states[state.name] = state
            # update the autoremove value if the state is defined as such
            if self.stateprops[state.name]['autoremove']:
                state.autoremove = True
            # set next and previous to the state itself (to ensure
            # that all states have a predecessor/successor)
            state.predecessor = state
            state.successor = state

        # handle default states
        self.states.is_default = False
        # if a default state is defined, store it and limit the list
        # of "available" states
        if default_state is not None:
            self.states[default_state].is_default = True
            self.states.default = self.states[default_state]
            self.states.available = (self.states.default,)
        else:
        # otherwise the list of available states is simply those which
        # are not autoremove
            self.states.default = None
            self.states.available = tuple(s for s in self.states if not s.autoremove)
        # print(self.states, self.states.default, self.states.available)

        # define "successor" and "predecessor" properties
        # start from the list of non-autoremove states
        usable = [s for s in self.states if not s.autoremove]
        for state in usable:
            # if the state has a "predecessor" state, use it
            if 'previous' in self.stateprops[state.name]:
                state.predecessor = self.states[self.stateprops[state.name]['previous']]
            else:
                # otherwise use the previous state in the list, the
                # first one being its own previous state
                state.predecessor = usable[max(0, usable.index(state) - 1)]
            # if the state has a "successor" state, use it
            if 'next' in self.stateprops[state.name]:
                state.successor = self.states[self.stateprops[state.name]['next']]
            else:
                # otherwise use the next state in the list, the
                # last one being its own next state
                state.successor = usable[min(len(usable)-1, usable.index(state) + 1)]

        # define function used from model to provide a random state
        self.model._values['_random_' + self.machine_name] = self.get_random_state
        # define function used from model to provide the default state
        # if any, otherwise a random state among non-autoremove ones
        if self.states.default:
            self.model._values['_default_' + self.machine_name] = self.get_default_state
        else:
            self.model._values['_default_' + self.machine_name] = self.get_random_state


    def get_random_state(self, caller=None):
        """Return a random state for this state machine."""
        return np.random.choice([state for state in self.states if not state.autoremove])

    def get_default_state(self, caller=None):
        """Return the default state for this state machine."""
        return self.states.default

    @property
    def state_colors(self):
        """Return a dictionary of state names associated with fill colors."""
        return {state.name: self._statedesc[state.name]['fillcolor']
                for state in self.states
                if not state.autoremove}


    def build_graph(self):
        """Parse the description of the state machine and extract the
        graph of the transitions between the states. Since a
        MultiDiGraph is used, each pair of nodes can be bound by
        several transitions if needed (beware the associated
        semantics).

        Example of YAML specification:
        ------------------------------
        transitions:
          - {from: S, to: I-, proba: p, cond: not_vaccinated}
          - {from: I-, to: S, proba: m}
          - {from: I-, to: I+m, proba: 'q*plp'}
          - {from: I-, to: I+, proba: 'q*(1-plp)'}

        """
        # add a node for each state
        for state in self.states:
            name = state.name
            self._statedesc[name]['tooltip'] = self.describe_state(name)
            self.graph.add_node(name, **self._statedesc[name])
        # build edges between states according to specified transitions
        if 'transitions' in self._description:
            self._parse_edges(self._description['transitions'],
                              type_id=enx.EdgeTypes.TRANSITION)
        if 'productions' in self._description:
            self._parse_edges(self._description['productions'],
                              type_id=enx.EdgeTypes.PRODUCTION)

    def _parse_edges(self, edges, type_id=enx.EdgeTypes.TRANSITION):
        """Parse the description of edges, with the difference
        transitions/productions

        """
        for edge in edges:
            from_ = edge['from']
            to_ = edge['to']
            others = {k: v for (k, v) in edge.items()
                      if k != 'from' and k != 'to'}
            for kwd in EDGE_KEYWORDS:
                if kwd in others:
                    # parm = pretty(sympify(others[kwd], locals=self.model._namespace))
                    parm = others[kwd]
                    label = '{}: {}'.format(kwd, parm)
            # label = ', '.join([pretty(sympify(x, locals=self.model._namespace))
            #                    for x in others.values()])
                    if str(parm) in self.model.parameters:
                        others['labeltooltip'] = self.model.describe_parameter(parm)
                    else:
                        others['labeltooltip'] = label
            # others['labeltooltip'] = ', '.join([self.model.describe_parameter(x)
            #                                     for x in others.values()
            #                                     if str(x) in self.model.parameters])
            # handle conditions if any on the edge
            cond, escape = None, False
            if 'cond' in others:
                cond = others['cond']
                others['truecond'] = others['cond']
            if ('escape' in others) and (type_id == enx.EdgeTypes.TRANSITION):
                cond = others['escape']
                escape = True
            if cond is not None:
                ### WARNING the operation below is not completely
                ### safe... it is done to replace conditions of the form
                ### 'x == y' by 'Eq(x, y)', but it is a simple
                ### substitution instead of parsing the syntax
                ### tree... Thus it is *highly* recommended to express
                ### conditions directly with Eq(x, y)
                if '==' in str(cond):
                    cond = 'Eq({})'.format(','.join(cond.split('==')))
                    # others['label'] = ', '.join(others.values())
            # if duration specified for this state, handle it as an
            # additional condition
            if ('duration' in self._statedesc[from_]) and (type_id == enx.EdgeTypes.TRANSITION):
                duration_cond = make_duration_condition(self.model, self.machine_name)
                if cond is None:
                    cond = duration_cond
                elif escape:
                    cond = 'AND(Not({}),{})'.format(duration_cond, cond)
                else:
                    cond = 'AND({},{})'.format(duration_cond, cond)
                    # print(cond)
                others['cond'] = cond
            if cond is not None:
                ## DEBUG:                print(cond, self.model._namespace)
                self.model.conditions[cond] = sympify(cond,
                                                      locals=self.model._namespace)
            # handle 'when' clause if any on the edge
            self._parse_when(others)
            # handle 'duration', 'escape' and 'condition' clauses if
            # any on the edge
            if type_id == enx.EdgeTypes.TRANSITION:
                self._parse_conditions_durations(from_, others)
            # parse actions on cross if any
            if ('on_cross' in others) and (type_id == enx.EdgeTypes.TRANSITION):
                l_actions = self._parse_action_list(others['on_cross'])
                others['actions'] = l_actions
            others['label'] = label
            others['type_id'] = type_id
            self.graph.add_edge(from_, to_, **others)
            # register rate/proba/amount expressions in the model
            for keyword in EDGE_KEYWORDS:
                if keyword in others:
                    self.model.add_expression(others[keyword])


    def _parse_when(self, edge_desc):
        """Parse the edge description in search for a 'when'
        clause. This special condition is aimed at globally assessing
        a time period within the whole simulation.

        """
        if 'when' in edge_desc:
            expression = sympify(edge_desc['when'],
                                 locals=self.model._event_namespace)
            edge_desc['when'] = str(expression)
            self.model._values[str(expression)] = make_when_condition(
                expression, modules=self.model.modules)

    def _parse_conditions_durations(self, from_, edge_desc):
        """Parse the edge description in search for durations,
        escapement and conditions specifications. Durations
        ('duration' clause )are handled as an additional condition
        (agents entering the state are given a 'time to live' in the
        state, then they are not allowed to leave the state until
        their stay reaches that value). Escapements ('escape' clause)
        are also translated as a condition, allowing the agent to
        leave the state when the expression is true, only while the
        stay duration is below its nominal value.

        """
        cond, escape = None, False
        if 'cond' in edge_desc:
            cond = edge_desc['cond']
        if 'escape' in edge_desc:
            cond = edge_desc['escape']
            escape = True
        if cond is not None:
            ### WARNING the operation below is not completely
            ### safe... it is done to replace conditions of the form
            ### 'x == y' by 'Eq(x, y)', but it is a simple
            ### substitution instead of parsing the syntax
            ### tree... Thus it is *highly* recommended to express
            ### conditions directly with Eq(x, y)
            if '==' in str(cond):
                cond = 'Eq({})'.format(','.join(cond.split('==')))
                # edge_desc['label'] = ', '.join(edge_desc.values()) if
        # duration specified for this state, handle it as an
        # additional condition
        if 'duration' in self._statedesc[from_]:
            duration_cond = make_duration_condition(self.model, self.machine_name)
            if cond is None:
                cond = duration_cond
            elif escape:
                cond = 'AND(Not({}),{})'.format(duration_cond, cond)
            else:
                cond = 'AND({},{})'.format(duration_cond, cond)
            edge_desc['cond'] = cond
        if cond is not None:
            self.model.conditions[cond] = sympify(cond, locals=self.model._namespace)

    def build_actions(self):
        """Parse the description of the state machine and extract the
        actions that agents running this state machine must have.

        Example of YAML specification:
        ------------------------------
        actions:
          say_hello:
            desc: action performed when entering the S state

        """
        for name, value in self._statedesc.items():
            for keyword in ['on_enter', 'on_stay', 'on_exit']:
                if keyword in value:
                    self._add_state_actions(name, keyword, value[keyword])
            if 'duration' in value:
                val = value['duration']
                self._add_state_duration_actions(name, val)

    def get_value(self, name):
        """Return the value associated with the specified name."""
        return self.model.get_value(name)


    def _add_state_duration_actions(self, state_name, duration_value):
        """Add implicit actions to manage stay duration in the specified state
        name. The `duration_value` can be either a parameter, a
        'statevar' or a distribution.

        """
        # initialize the actions associated to the state if none
        if state_name not in self.state_actions:
            self.state_actions[state_name] = {}
        # retrieve the list of actions on enter for this state, if any
        lenter = self.state_actions[state_name]['on_enter']\
                   if 'on_enter' in self.state_actions[state_name] else []
        # build a partial function based on the current state machine name
        enter_action = partial(make_duration_init_action,
                               machine_name=self.machine_name)
        # set the name of the action
        enter_action.__name__ = 'init_duration'
        # set the action parameters (the expression associated to the duration)
        enter_params = [self.model.add_expression(duration_value)]
        # instantiate the action
        init_action = AbstractAction.build_action('duration',
                                                  function=enter_action,
                                                  l_params=enter_params,
                                                  state_machine=self)
        # and insert it at the beginning of the list of actions
        lenter.insert(0, init_action)
        self.model.add_init_action(self.machine_name,
                                   self.states[state_name],
                                   init_action)
        # lstay = self.state_actions[name]['on_stay']\
        #           if 'on_stay' in self.state_actions[name] else []
        # stay_action = partial(make_TTL_increase_action,
        #                       machine_name=self.machine_name)
        # stay_action.__name__ = '+_time_spent'
        # lstay.insert(0, AbstractAction.build_action('duration',
        #                                             function=stay_action,
        #                                             state_machine=self))
        self.state_actions[state_name]['on_enter'] = lenter
        # self.state_actions[name]['on_stay'] = lstay

    def _add_state_actions(self, name, event, actions):
        """Add the specified actions for the state with the given
        name, associated with the event (e.g. 'on_stay', 'on_enter',
        'on_exit'). Expressions contained in the parameters lists or
        dicts are automatically expanded.

        """
        if name not in self.state_actions:
            self.state_actions[name] = {}
        l_actions = self._parse_action_list(actions)
        self.state_actions[name][event] = l_actions

    def _parse_action_list(self, actions):
        """Parse the list of actions associated with a state."""
        l_actions = []
        for d_action in actions:
            if 'action' in d_action:
                action = d_action['action']
                l_params = [self.model.add_expression(expr)
                            for expr in d_action['l_params']]\
                                if 'l_params' in d_action\
                                else []
                # if 'd_params' in d_action:
                #     print(d_action['d_params'])
                d_params = {key: self.model.add_expression(expr)
                            for key, expr in d_action['d_params'].items()}\
                                if 'd_params' in d_action\
                                else {}
                l_actions.append(
                    AbstractAction.build_action('action',
                                                method=action,
                                                l_params=l_params,
                                                d_params=d_params,
                                                state_machine=self))
            else: #TODO: dispatch through AbstractAction (factory),
                  #make subclasses responsible for parameter parsing
                understood = False
                for keyword in ['increase', 'decrease',
                                'increase_stoch', 'decrease_stoch']:
                    if keyword in d_action:
                        # assume that increase statevar with rate
                        l_actions.append(
                            AbstractAction.build_action(
                                keyword,
                                statevar_name=d_action[keyword],
                                parameter=self.model.add_expression(d_action['rate']),
                                delta_t=self.model.delta_t,
                                state_machine=self
                            )
                        )
                        understood = True
                for keyword in ['set_var']:
                    if keyword in d_action:
                        # assume that increase statevar with rate
                        l_actions.append(
                            AbstractAction.build_action(
                                keyword,
                                statevar_name=d_action[keyword],
                                parameter=self.model.add_expression(d_action['value']),
                                model=self.model
                            )
                        )
                        understood = True
                for keyword in ['become', 'clone', 'produce_offspring']:
                    if keyword in d_action:
                        amount = d_action['amount'] if 'amount' in d_action else None
                        probas = d_action['proba'] if 'proba' in d_action else None
                        l_actions.append(
                            AbstractAction.build_action(
                                keyword,
                                prototypes=d_action[keyword],
                                amount = amount,
                                probas = probas,
                                model = self.model
                            )
                        )
                        understood = True
                for keyword in ['log_vars']:
                    if keyword in d_action:
                        vars = [self.model.add_expression(varname)
                                for varname in d_action[keyword]]
                        l_actions.append(
                            AbstractAction.build_action(
                                keyword,
                                parameter=None,
                                l_params=vars
                            )
                        )
                        understood = True
                for keyword in ['message', 'record_change']:
                    if keyword in d_action:
                        l_actions.append(
                            AbstractAction.build_action(
                                keyword,
                                parameter=d_action[keyword]
                            )
                        )
                        understood = True
                if not understood:
                    print('ERROR !!!!') # but there is certainly a fatal error !
                    print(d_action)
        return l_actions




    #----------------------------------------------------------------
    # Output facilities

    def describe_state(self, name):
        """Return the description of the state with the specified
        name.

        """
        desc = self._statedesc[name]
        return "{} ({}):\n\t{}".format(name, desc['name'], desc['desc'])

    def write_dot(self, filename, view_actions=True):
        """Write the graph of the current state machine in the
        specified filename, according to the dot/graphviz format.

        """

        rankdir = "LR" if self.graph.edges() else "TB"
        output = '''digraph {
          charset="utf-8"
        '''
        output += '''\trankdir={};
        '''.format(rankdir)
        output += '''
        \tnode[fontsize=16, fontname=Arial, shape=box, style="filled,rounded"];
        \tedge[minlen=1.5, fontname=Times, penwidth=1.5, tailtooltip="", headtooltip=""];

        '''
        for state in self.states:
            name = state.name
            name_lab = name
            if 'duration' in self._statedesc[name]:
                name_lab += '&nbsp;{}'.format(CLOCK_SYMBOL)
            actions = 'shape="Mrecord", label="{}", '.format(name_lab)
            nodestyle = "filled,rounded"
            if state.is_default:
                nodestyle += ",bold"
            if state.autoremove:
                nodestyle += ",dotted"
            if view_actions:
                onenter = ACTION_SYMBOL+'|'\
                          if 'on_enter' in self._statedesc[name] else ''
                onstay = '|'+ACTION_SYMBOL\
                         if 'on_stay' in self._statedesc[name] else ''
                onexit = '|'+ACTION_SYMBOL\
                         if 'on_exit' in self._statedesc[name] else ''
                if onenter or onstay or onexit:
                    actions = 'shape="Mrecord", label="{%s{\ %s\ %s}%s}", ' % (
                        onenter, name_lab, onstay, onexit)
            output += '\t"{}" [{}tooltip="{}", fillcolor={}, style="{}"] ;\n'.format(
                name, actions,
                self._statedesc[name]['tooltip'],
                # '\n\tON ENTER: {}'.format(self.state_actions[name]['on_enter'])\
                # if onenter else '' +\
                # '\n\tON STAY: {}'.format(self.state_actions[name]['on_stay'])\
                # if onstay else '' +\
                # '\n\tON EXIT: {}'.format(self.state_actions[name]['on_exit'])\
                # if onexit else '',
                self._statedesc[name]['fillcolor'],
                nodestyle)
        for from_, to_ in SortedSet(self.graph.edges()):
            for desc in self.graph.edge[from_][to_].values():
                edgetip = ''
                tail = 'none'
                if 'when' in desc:
                    tail += WHEN_SYMBOL
                    edgetip += 'WHEN: {}'.format(desc['when'])
                if 'escape' in desc:
                    tail += ESCAPE_SYMBOL
                    edgetip += 'ESCAPE: {}'.format(desc['escape'])
                if 'truecond' in desc:
                    tail += COND_SYMBOL
                    edgetip += 'COND: {}'.format(desc['truecond'])
                head = 'normalnone'
                if 'on_cross' in desc:
                    head += CROSS_SYMBOL
                    # edgetip += 'ON CROSS: {}\\n'.format(desc['on_cross'])
                output += ('\t"{}" -> "{}" [label="{}", labeltooltip="{}", '
                           'arrowtail="{}", arrowhead="{}", dir=both, '
                           'tooltip="{}", minlen=3, style="{}"];\n').format(
                               from_, to_, desc['label'], desc['labeltooltip'],
                               tail, head, edgetip, desc['type_id'].linestyle)
        output += '}'
        with open(filename, 'w', encoding="utf8") as f:
            f.write(output)



"""A Python implementation of the EMuLSion framework (Epidemiologic
MUlti-Level SImulatiONs).

Classes and functions for entities management.
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from   collections               import OrderedDict
import numpy                     as np

from   sortedcontainers          import SortedSet

from   emulsion.agent.core       import Aggregation
from   emulsion.tools.misc       import select_random, AGENTS

#   _____ _                 _   __      ___
#  / ____(_)               | |  \ \    / (_)
# | (___  _ _ __ ___  _ __ | | __\ \  / / _  _____      __
#  \___ \| | '_ ` _ \| '_ \| |/ _ \ \/ / | |/ _ \ \ /\ / /
#  ____) | | | | | | | |_) | |  __/\  /  | |  __/\ V  V /
# |_____/|_|_| |_| |_| .__/|_|\___| \/   |_|\___| \_/\_/
#                    | |
#                    |_|

class SimpleView(Aggregation):
    """A SimpleView uses a set to store the underlying
    units. It is rather aimed at storing AtomAgents. All units are
    considered in the same state, thus conditions in the state
    machines are evaluated for the whole compartment, not for
    individuals.

    """
    def __init__(self, **others):
        super().__init__(**others)
        self._content = SortedSet()
        self.stochastic = True

    def __iter__(self):
        return self._content.__iter__()

    def get_content(self):
        """Return the units contained in the current unit.

        """
        return ('agents', list(self._content))

    def add(self, population):
        """Add the specified population to the current compartment."""
        super().add(population)
        self._content |= SortedSet(population)

    def remove(self, population):
        """Remove the specified population from the current
        compartment.

        """
        super().remove(population)
        self._content -= SortedSet(population)

    def next_states(self, states, values, populations, actions, method=None):
        """Compute stochastically the population moving from the current
        compartment to each of the destination states, according to
        the values interpreted as probabilities. Actions are to be
        performed when changing state. The population affected by the
        transitions is stored in the first element of the
        `populations` parameter, as a tuple: ('agents',
        list_of_agents). Several edges can lead to the same state.

        Return a list of tuples:
          (state, {'agent': list of agents, 'actions': list of actions})
        """
        # retrieve the list of agents from the tuple at first position
        origin = populations[0][AGENTS]
        if method == 'amount':
            # length of values is expected to be the number of output edges
            total_value = sum(values)
            if total_value > len(origin):
                # restart with proportions instead
                return self.next_states(states,
                                        tuple(v / total_value for v in values) + (0,),
                                        populations, actions, method=None)
            evolution = [int(v) for v in values]
            # print(states, evolution)
        else:
            # values is expected to be the number of output edges + 1
            evolution = np.random.multinomial(len(origin), values)[:-1]

        ### now select agents corresponding to the amounts
        # init change list
        changes = []
        # track agents already selected
        modified = SortedSet()
        for state, qty, act in zip(states[:-1], evolution, actions[:-1]):
            if qty > 0:
                # pick up agents randomly (not already chosen) in the
                # origin group
                moving = select_random(origin, qty, exclude=modified)
                modified |= SortedSet(moving)
                # append this subgroup to the change list
                changes.append((self._host.state_machine.states[state],
                              {'agents': moving, 'actions': act}))
        return changes



#              _             _   _        __      ___
#     /\      | |           | | (_)       \ \    / (_)
#    /  \   __| | __ _ _ __ | |_ ___   ____\ \  / / _  _____      __
#   / /\ \ / _` |/ _` | '_ \| __| \ \ / / _ \ \/ / | |/ _ \ \ /\ / /
#  / ____ \ (_| | (_| | |_) | |_| |\ V /  __/\  /  | |  __/\ V  V /
# /_/    \_\__,_|\__,_| .__/ \__|_| \_/ \___| \/   |_|\___| \_/\_/
#                     | |
#                     |_|

class AdaptiveView(SimpleView):
    """An AdaptiveView is able to evaluate conditions on individuals
    (AtomAgents) for state transitions, and to detect changes in its
    content and automatically report it to the upper level. Each
    compartment is associated with a particular value of a specific
    state variable or attribute of the units in its content. At the
    end of the evolve step, the compartment checks the values of the
    units in its content and reports units where changes occured.

    """
    def __init__(self, observables=(None,), values=(None,), **others):
        """Specify the state variables or attributes associated with
        this compartment (``observables``) and the expected values.

        """
        super().__init__(**others)
        self.statevars.observables = observables
        for (observable, value) in zip(observables, values):
            self.statevars[observable] = value


    def evaluate_condition(self, condition):
        """Return the population of the compartment if the condition
        is fulfilled, 0 otherwise.

        """
        if not callable(condition):
            return self.get_content() if condition else ()
        result = list(filter(condition, self._content))
        return ('agents', result) if result else ()

    def next_states(self, states, values, populations, actions, method=None):
        """Compute stochastically the population moving from each population
        (tuples ('agents': list_of_agents)) in the `populations`
        parameter, to each of the possible destination states,
        according to each of the values interpreted as
        probabilities. Actions are to be performed when changing
        state. Several edges can lead to the same state.

        Return a list of tuples:
          (state, {'agent': list of agents, 'actions': list of actions})
        """
        if len(populations) == 1:
            if method == 'amount':
                # length of values is expected to be the number of output edges
                total_value = sum(values)
                if total_value > len(populations[0][AGENTS]):
                    # restart with proportions instead
                    return self.next_states(states,
                                            tuple(v / total_value
                                                  for v in values) + (0,),
                                            populations, actions, method=None)
                values = values + (total_value,)
            populations *= len(states)-1
        # init change list
        changes = []
        # track agents already selected
        modified = SortedSet()
        for state, value, population, act\
          in zip(states[:-1], values[:-1], populations, actions[:-1]):
            if method == 'amount':
                evolution = int(value)
            else:
                evolution = np.random.binomial(len(population[AGENTS]), value)
            if evolution > 0:
                ### now select agents corresponding to the amounts
                # pick up agents randomly (not already chosen) in the
                # origin group
                moving = select_random(population[AGENTS],
                                       evolution, exclude=modified)
                modified |= SortedSet(moving)
                # append this subgroup to the change list
                changes.append((self._host.state_machine.states[state],
                              {'agents': moving, 'actions': act}))
        return changes


    def evolve(self, machine=None):
        """After the ordinary ``evolve`` step, check units which have
        changed their value(s) of the compartment-specific
        observable(s). The corresponding units are reported to the
        upper level.

        """
        super().evolve(machine=machine)
        self.check_consistency()

    def check_consistency(self):
        """Check the value of the observables of the current
        compartment. Units which do not have the expected values are
        notified to the upper level.

        """
        # intruders = [unit for unit in self
        #              if any(unit.get_information(observable)\
        #                       != self.statevars[observable]
        #                     for observable in self.statevars.observables)]
        intruders = SortedSet()
        for unit in self:
            for observable in self.statevars.observables:
                if unit.get_information(observable) != self.statevars[observable]:
                    intruders.add(unit)
                    break
        if intruders:
            self._host.notify_changed_units(self, intruders)

    def clone(self, **others):
        """Make a copy of the current compartment with the specified
        observable/value settings. The new content is empty.

        """
        # new_comp = copy.copy(self)
        # new_comp._content = SortedSet()
        # new_comp.statevars = StateVarDict(self.statevars)
        new_comp = self.__class__.from_dict(self.statevars)
        new_comp.model = self.model
        new_comp._content = SortedSet()
        new_comp._host = self._host
        new_comp.statevars.update(**others)
        # new_comp.statevars.step = self.statevars.step
        new_comp.recursive = self.recursive
#        new_comp.set_information(observable, value)
        return new_comp



#   _____ _                   _                      ___      ___
#  / ____| |                 | |                    | \ \    / (_)
# | (___ | |_ _ __ _   _  ___| |_ _   _ _ __ ___  __| |\ \  / / _  _____      __
#  \___ \| __| '__| | | |/ __| __| | | | '__/ _ \/ _` | \ \/ / | |/ _ \ \ /\ / /
#  ____) | |_| |  | |_| | (__| |_| |_| | | |  __/ (_| |  \  /  | |  __/\ V  V /
# |_____/ \__|_|   \__,_|\___|\__|\__,_|_|  \___|\__,_|   \/   |_|\___| \_/\_/


class StructuredView(Aggregation):
    """A StructuredView uses ad dict list to store the underlying
    units. It is rather aimed at storing other compartiments.

    """
    def __init__(self, keep_history=True, **others):
        super().__init__(**others)
        self.keep_history = keep_history
        self.new_population = []
        self._content = OrderedDict()
        self._notifications = {}

    def __iter__(self):
        return self._content.values().__iter__()

    def __getitem__(self, name):
        if name not in self._content: # DEBUG: should not happen !
            print(name)
        return self._content[name]

    def get_content(self):
        """Return the agents contained in the current view.

        """
        print('WARNING, should not enter here !!!')
        return ('agents', list(self._content.values()))

    def add(self, population):
        """Add the specified population to the current
        compartment. The population is expected to be a dictionary
        with names as keys and compartments as values.

        """
        for key, pop in population.items():
            if key in self._content:
                self._content[key].add(pop._content)
            else:
                self._content[key] = pop

    def remove(self, population):
        """Remove the specified population from the current
        compartment.  The population is expected to be a dictionary
        with names as keys and compartments as values.

        """
        for key, pop in population.items():
            if key in self._content:
                self._content[key].remove(pop._content)

    def notify_changed_units(self, source_compartment, units):
        """Receive a change notification concerning several units
        which cannot be stored anymore in the source compartment.

        """
        if source_compartment in self._notifications:
            self._notifications[source_compartment] |= units
        else:
            self._notifications[source_compartment] = SortedSet(units)

    def handle_notifications(self):
        """Handle all notifications received during the time step."""
        for source_compartment, units in self._notifications.items():
            source_observables = source_compartment.statevars.observables
            new_pop = {}
 #           pop_values = {}
            for unit in units:
                key = tuple(unit.get_information(k) for k in source_observables)
                if key not in new_pop:
                    new_pop[key] = [unit]
#                    pop_values[key] = unit.get_information(source_observable)
                else:
                    new_pop[key].append(unit)
            for key, unit in new_pop.items():
                # print(source_observables, key)
                new_comp = self.get_or_build(key, source=source_compartment)
#                print('Moving to:', new_comp, '\n\t', [u._agid for u in unit])
                source_compartment.move_to(new_comp, agents=unit)
        self._notifications.clear()

    def get_or_build(self, key, source=None):
        """Return the compartment with the specified key if any, or
        build one by cloning the source if not found.

        """
        if key not in self._content:
            key_infos = dict(zip(self.keys, key))
            # print('Adding new comp:', key, key_infos)
            new_compart = source.clone(host=self, **key_infos)
            new_compart.keys = key
            # if any key for compart is a state marked "autoremove",
            # then the whole compart is marked "autoremove"
            new_compart.autoremove = any(hasattr(value, 'state_machine') and\
                                  value.state_machine.get_property(value.name, 'autoremove')
                                  for value in key)
            # print(new_compart, new_compart.autoremove)
            self.add({key: new_compart})
        return self[key]

    def evolve(self, machine=None):
        super().evolve(machine=machine)
        self.handle_notifications()




class StructuredViewWithCounts(StructuredView):
    """Same as StructuredView, plus the capability of handling counts for
    key values.

    """
    def __init__(self, allowed_values=None, **others):
        super().__init__(**others)
        # if allowed_values:
        self.allowed_values = allowed_values
        self.init_counts()

    def init_counts(self, index=0):
        """Initialize the counts."""
        self.counts = {}
        if self.allowed_values:
            self.counts = {value: [] if self.keep_history else 0
                           for value in self.allowed_values}
        self.counts['step'] = [] if self.keep_history else 0

    def update_counts(self, index=0):
        """Update the counts.

        """
        if self.allowed_values:
            total = {value: 0 for value in self.allowed_values}
            for (key, unit) in self._content.items():
                if key[index] is not None and key[index] in self.allowed_values:
                    total[key[index]] += unit.get_information('population')
            if self.keep_history:
                self.counts['step'].append(self.statevars.step)
                for value in self.allowed_values:
                    self.counts[value].append(total[value])
            else:
                self.counts['step'] = self.statevars.step
                self.counts.update(total)

    def __contains__(self, keys):
        return keys in self._content

    def evolve(self):
        super().evolve()
        self.update_counts()




#                _        _____ _                   _                      _
#     /\        | |      / ____| |                 | |                    | |
#    /  \  _   _| |_ ___| (___ | |_ _ __ _   _  ___| |_ _   _ _ __ ___  __| |
#   / /\ \| | | | __/ _ \\___ \| __| '__| | | |/ __| __| | | | '__/ _ \/ _` |
#  / ____ \ |_| | || (_) |___) | |_| |  | |_| | (__| |_| |_| | | |  __/ (_| |
# /_/    \_\__,_|\__\___/_____/ \__|_|   \__,_|\___|\__|\__,_|_|  \___|\__,_|
# __      ___
# \ \    / (_)
#  \ \  / / _  _____      __
#   \ \/ / | |/ _ \ \ /\ / /
#    \  /  | |  __/\ V  V /
#     \/   |_|\___| \_/\_/


class AutoStructuredView(StructuredView):
    """An AutoStructuredView stores agents in an OrderedDict, using a
    specific statevar as key.

    """
    def __init__(self, key_variable=None, **others):
        super().__init__(**others)
        self.key_variable = key_variable
        self.current_key = 0    # auto-increment ID if none specified


    def add(self, population):
        """Add the specified population (SortedSet or list) to the
        current view.

        """
        for agent in population:
            if self.key_variable not in agent.statevars:
                agent.statevars[self.key_variable] = self.current_key
                self.current_key += 1
            else:
                self.current_key = max(1 + agent.statevars[self.key_variable],
                                       self.current_key)
            self._content[agent.statevars[self.key_variable]] = agent

    def remove(self, population):
        """Remove the specified population from the current view. The
        population is expected to be a SortedSet or list.

        """
        for agent in population:
            key = agent.statevars[self.key_variable]
            if key in self._content:
                del self._content[key]



""".. module:: emulsion.tools.calendar

Classes and functions for the definition of Emulsion calendars.
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from   sortedcontainers      import SortedDict


#  ______                    _   _
# |  ____|                  | | (_)
# | |__  __  _____ ___ _ __ | |_ _  ___  _ __  ___
# |  __| \ \/ / __/ _ \ '_ \| __| |/ _ \| '_ \/ __|
# | |____ >  < (_|  __/ |_) | |_| | (_) | | | \__ \
# |______/_/\_\___\___| .__/ \__|_|\___/|_| |_|___/
#                     | |
#                     |_|

class InvalidIntervalException(Exception):
    """Exception raised when trying to insert an inconsistent event in the
    calendar. An event is considered inconsistent if the begin date is
    posterior to the send date in a non-periodic calendar.

    """
    def __init__(self, begin, end):
        """Create the exception with the incorrect *begin* and *end* dates."""
        super().__init__()
        self.value = (begin, end)

    def __str__(self):
        return 'Invalid dates interval {}'.format(self.value)


#  ______                _   _
# |  ____|              | | (_)
# | |__ _   _ _ __   ___| |_ _  ___  _ __  ___
# |  __| | | | '_ \ / __| __| |/ _ \| '_ \/ __|
# | |  | |_| | | | | (__| |_| | (_) | | | \__ \
# |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/

def date_in(begin, end, period=None):
    """Build a function which tests if a date belongs to the specified
    interval (from *begin* to *end*). If a *period* is specified, the
    test relies upon the periodicity, otherwise dates are considered
    absolute.

    Parameters
    ----------
    begin: ``datetime.datetime``
        the date when the interval begins
    end: ``datetime.datetime``
        the date when the interval ends
    period: ``datetime.timedelta``
        the duration of the period after which the calendar cycles

    Returns
    -------
    lambda:
        a function which maps a date (``datetime.datetime``) to a
        ``bool`` to indicate whether or not the date belongs to the
        *begin*-*end* interval.

    Raises
    ------
    InvalidIntervalexception
        if *begin* is posterior to *end* in a non-periodic calendar
    """
    if period is not None:
        begin, end = begin % period, end % period
        if begin <= end:
            def test_date(date):
                """Check that the specified date belongs to the event interval,
                considering the periodicity.

                """
                return begin <= (date % period) <= end
        else:
            def test_date(date):
                """Check that the specified date belongs to the event interval,
                considering the periodicity and the fact that the end
                date appears before the begin date in a civil year.

                """
                return (date % period) <= end or (date % period) >= begin
    else:
        if begin <= end:
            def test_date(date):
                """Check that the specified date belongs to the event interval,
                considered as absolute (no periodicity).

                """
                return begin <= date <= end
        else:
            raise InvalidIntervalException(begin, end)
    return test_date


#  ______               _    _____      _                _
# |  ____|             | |  / ____|    | |              | |
# | |____   _____ _ __ | |_| |     __ _| | ___ _ __   __| | __ _ _ __
# |  __\ \ / / _ \ '_ \| __| |    / _` | |/ _ \ '_ \ / _` |/ _` | '__|
# | |___\ V /  __/ | | | |_| |___| (_| | |  __/ | | | (_| | (_| | |
# |______\_/ \___|_| |_|\__|\_____\__,_|_|\___|_| |_|\__,_|\__,_|_|


class EventCalendar:
    """The EventCalendar class is intended to handle events and
    periods of time. Dates/times are given to the calendar in a
    human-readable format, and converted to integers (simulation
    steps).

    """
    def __init__(self, calendar_name, step_duration, origin,
                 period=None, **initial_dates):
        """Initialize the calendar using the specified information.

        Parameters
        ----------
        calendar_name: string
            the name of the calendar
        step_duration: ``datetime.timedelta``
            actual duration of one simulation step
        origin: ``datetime.datetime``
            date/time value of the beginning of the calendar
        period: ``datetime.timedelta``
            if the calendar is periodic, actual duration of the
            period, None otherwise
        initial_dates: dict
             initial dictionary of events (event name as key, dict as
             value with either begin/end dates or one single date if
             punctual event)

        """
        self.calendar_name = calendar_name
        self.step = 0
        self.step_duration = step_duration
        self.origin = origin
        self.current_date = self.origin
        self.events = SortedDict()
        self.description = SortedDict()
        self.period = None if period is None\
          else period // self.step_duration
        for name, value in initial_dates.items():
            self.add_event(name, value)


    def increment(self, steps=1):
        """Advance the current date by the specified number of
        simulation steps.

        Parameters
        ----------
        steps: int
            number of steps to 'add' to the current date
        """
        self.current_date += self.step_duration * steps

    def step_to_date(self, step):
        """Return the date when the specified step begins.

        Parameters
        ----------
        step: int
            the time step to convert to a date

        Returns
        -------
        d: ``datetime.datetime``
            the corresponding date
        """
        return self.origin + self.step_duration * step

    def date_to_step(self, date):
        """Return the step corresponding the specified date.

        Parameters
        ----------
        date: ``datetime.datetime``
            the date to convert into time steps

        Returns
        -------
        t: int
            the time step during which the specified date occurs (the
            date occurs between the returned time step *t* and time
            step *t+1*)

        """
        return (date - self.origin) // self.step_duration


    def add_event(self, name, begin_end):
        """Add the specified event to the calendar. An event is characterized
        by its *name* and a *begin_end* tuple indicating the begin and
        end dates.

        Parameters
        ----------
        name: str
            the name of the event
        begin_end: tuple
           a tuple composed of the begin date and the end date
           (possibly identical for handling punctual events)

        """
        begin_date, end_date = begin_end
        begin = self.date_to_step(begin_date)
        end = self.date_to_step(end_date)
        if end < begin and self.period is None:
            raise InvalidIntervalException(begin, end)
        self.description[name] = begin_end
        self.events[name] = date_in(begin, end, self.period)

    def get_events(self):
        """Return the list of events contained in the current calendar.

        Returns
        -------
        l: list
            the list of all event names
        """
        return list(self.description.keys())

    def __getitem__(self, name):
        return self.events[name]



"""Usage:
    emulsion run [--plot] MODEL [options] [(-p KEY=VALUE)...]
    emulsion show MODEL [options] [(-p KEY=VALUE)...]
    emulsion describe MODEL PARAM...
    emulsion diagrams MODEL [options]
    emulsion plot MODEL [options]
    emulsion check MODEL [options]
    emulsion generate MODEL
    emulsion (-h | --help | -V | --version | -L | --license)

Commands:
  run MODEL                   Run simulations based on the specified MODEL
                              (MODEL is the path to the YAML file describing the
                              model to run).
  show MODEL                  Print all MODEL parameter values and exit.
  describe MODEL PARAM...     Describe the role of specified PARAMeters in the
                              MODEL and exit.
  diagrams MODEL              Produce model diagrams (as option --view-model
                              when running/plotting) and open them
  plot MODEL                  Plot outputs for MODEL (assumed already run) and
                              exit.
  check MODEL                 Check the syntactic correctness of the given MODEL
                              (path to YAML file of the model to run), according
                              to EMULSION meta-model. Option '--meta-tree'
                              generates a figure representing EMULSION meta-
                              model. If the model is correct, option
                              '--model-tree' also generates a figure for the
                              MODEL structure.
  generate MODEL              Generate a skeleton to help writing specific
                              pieces of code before the MODEL can be run, and
                              exit.

Options:
  -h --help                   Display this page and exit.
  -V --version                Display version number and exit.
  -L --license                Display license and exit.
  --plot                      Plot outputs just after running the model.
  -r RUNS --runs RUNS         Specify the number of repetitions of the same
                              model [default: 10].
  -t STEPS --time STEPS       Specify the number of time steps to run in each
                              repetition. If the model defines a total_duration,
                              it is used as time limit, unless the '-t' option
                              is explicitly specified. Otherwise, the default
                              value is 100 steps.
  -p KEY=VAL --param KEY=VAL  Change parameter named KEY to the specified VALue.
  --view-model                Produce diagrams to represent the state machines
                              of the model (requires Graphviz). Figures are
                              stored in figure-dir.
  --silent                    Show only the progression of repetitions instead
                              of the progression of each simulation.
  --quiet                     Show no progression information at all.
  --save FILE                 Save simulation state (all agents state and
                              parameters) at the end of the simulation.
  --load FILE                 Use a saved simulation to start the current one.
  --output-dir OUTPUT         Specify a directory for simulation outputs
                              [default: outputs].
  --input-dir INPUT           Specify a directory for simulation inputs
                              [default: data].
  --figure-dir FIG            Specify a directory for graphic outputs (figures)
                              [default: img].
  --log-params                When producing CSV outputs, insert the name and
                              value of each parameter explicitly changed by
                              option -p/--param.
  --format FORMAT             Specify an image format for diagram outputs
                              [default: svg].

Advanced options:
  --seed SEED                 Set a seed value for random numbers. When not
                              specified, the seed is set according to system
                              time and the process id.
  --show-seed                 Print the seed used for random numbers.
  --start-id ID               ID of the first repetition of the same model
                              [default: 0].
  --echo                      Just print command-line arguments parsed by Python
                              docopt module and exit.
  --deterministic             Run the simulation in deterministic mode if
                              available.
  --modifiable                Output the list of modifiable parameters and exit.
  --level LEVEL               Specify the LEVEL (scale) for running the model.
                              Valid values are those defined in the 'level'
                              section of the MODEL. The corresponding agent
                              class will be used to manage the simulation of
                              lower-level entities. When no value is given, the
                              highest level (if any) is used.
  --model-tree                Output a figure or the syntactic tree that
                              represents the model if it complies to EMULSION
                              DSL syntax (requires Graphviz).
  --meta-tree                 Output a figure of the meta-model associated with
                              EMULSION DSL (requires Graphviz).


EMULSION (Epidemiological Multi-Level Simulation framework)
===========================================================

Contributors and contact:
-------------------------

    - Sébastien Picault (sebastien.picault@inrae.fr)
    - Yu-Lin Huang
    - Vianney Sicard
    - Sandie Arnoux
    - Gaël Beaunée
    - Pauline Ezanno (pauline.ezanno@inrae.fr)

    INRAE, Oniris, BIOEPAR, 44300, Nantes, France


How to cite:
------------

    S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
    P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
    stochastic models in human, animal and plant epidemiology", PLoS
    Computational Biology 15(9): e1007342. DOI:
    10.1371/journal.pcbi.1007342


License:
--------

   Copyright 2016 INRAE and Univ. Lille

   Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000

   Agence pour la Protection des Programmes,
   54 rue de Paradis, 75010 Paris, France

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

## TO BE IMPLEMENTED LATER:

  #   emulsion sensitivity MODEL DESIGN [options]
  #   emulsion change MODEL NEW_MODEL (-p KEY=VALUE)...


  # sensitivity MODEL DESIGN    Run a sensitivity analysis for the specified MODEL
  #                             using the provided experimental DESIGN.
  #                             NOT IMPLEMENTED YET.
  # change MODEL NEW_MODEL      Modify the initial MODEL into a NEW_MODEL using
  #                             appropriate options, and exit. NOT IMPLEMENTED YET

  # --init INIT_FILE            Speficy a file for initial conditions.
  #                             NOT USED YET.
  # --test                      Run the simulation in test mode. NOT USED YET.
  # --metamodel META            Specify a metamodel for syntax checking
  #                             [default: ../../scripts/emulsion.tx]


import sys
import os
import time
import subprocess
import datetime                     as dt
import webbrowser

from   pathlib                      import Path

import yaml
import numpy                        as np

from   docopt                       import docopt
from   textx                        import metamodel_from_file
from   textx.export                 import metamodel_export, model_export
import colorama

import emulsion
from   emulsion.model               import EmulsionModel
from   emulsion.tools.state         import StateVarDict
from   emulsion.tools.misc          import load_class
from   emulsion.tools.plot          import plot_outputs
from   emulsion.tools.simulation    import MultiSimulation

VERSION = '1.0.12'
DEFAULT_VERSION = "1.0"
LICENSE = 'Apache-2.0'
DEFAULT_LICENSE = "Apache-2.0"

def get_metamodel():
    """Return the path to the metamodel used for checking syntax of
    EMULSION YAML files.

    """
    # retrieve path to EMULSION repository
    parts = Path(emulsion.__file__).parts[:-1] + ("resources", "emulsion.tx")
    return Path(*parts)

def get_version():
    """Retrieve the version number of current program.

    """
    # try:
    #     proc = subprocess.Popen(["git", "describe",
    #                              "--tags", "--dirty", "--always"],
    #                             stdout=subprocess.PIPE)
    # except EnvironmentError:
    #     print("unable to run git")
    #     return 'Unknown'
    # stdout = proc.communicate()[0].strip().decode('utf-8')
    # if proc.returncode != 0:
    #     print("unable to run git")
    #     return DEFAULT_VERSION
    # return stdout
    return DEFAULT_VERSION if VERSION.startswith('[') else VERSION

def get_license():
    """Retrieve the license  of current program.

    """
    return DEFAULT_LICENSE if LICENSE.startswith('[') else LICENSE

def change_parameters(params, change_list):
    """Change either the model or local parameters according to the list
    of new values.

    """
    model_changes = {}
    modifiable = params.model.get_modifiable_parameters()
    for key, val in [p.split('=') for p in change_list]:
        if key in params:
            # if the new value is the name of another parameter,
            # retrieve its current value
            if val in params:
                val = params[val]
            # retrieve the value with correct type
            params[key] = type(params[key])(val)
        elif key in modifiable:
            model_changes[key] = val
        else:
            print(colorama.Fore.RED + colorama.Style.BRIGHT +\
                  'Unknown parameter:{}'.format(key) + colorama.Style.RESET_ALL)
            sys.exit(-1)
    if model_changes:
        params.model.change_parameter_values(model_changes,
                                             log_params=params.log_params)

def describe_parameters(params):
    """Display the role of all parameters specified in the PARAM argument
    and exit.

    """
    model = params.model
    print(colorama.Style.BRIGHT + '\n{!s: ^72}'.format(params.model))
    print('{: ^72}'.format('ROLE OF PARAMETERS (AND CURRENT DEFINITION)') +\
          colorama.Style.RESET_ALL)
    print('-'*72)
    for name in params.to_describe:
        print(model.describe_name(name))
    print('-'*72)


def show_parameters(params, short=False):
    """Display all parameters involved in the current program and model
    and exit.

    """
    modifiable = params.model.get_modifiable_parameters()
    if short:
        print(' '.join('{}={}'.format(key, params.model.get_value(key))
                       for key in modifiable))
        sys.exit()
    print(colorama.Style.BRIGHT + '\n{!s: ^72}'.format(params.model))
    print('{: ^72}'.format('AVAILABLE PARAMETERS (with their current value)'))
    print('-'*72)
    print('MODEL PARAMETERS')
    print('-'*72 + colorama.Style.RESET_ALL)
    for key, val in modifiable.items():
        print('  {:.<34}{!s:.>34}'.format(key, val))
    print('-'*72)
    # print('PROGRAM PARAMETERS')
    # print('-'*72)
    # for key, val in params.items():
    #     print('  {:.<34}{!s:.>34}'.format(key, val))
    sys.exit()


def generate_model(params):
    """Generate a skeleton for the pieces of specific code to write. If
    needed, create subdirectories. If files already exist, add a timestamp
    to the filename.

    """
    model = params.model
    src_path = Path(__file__).parent.parent
    paths = sorted(set([Path(level_desc['file'])
                        for level_desc in model.levels.values()
                        if 'file' in level_desc
                        # if not level_desc['module'].startswith('emulsion.agent')
    ]))
    for mod_path in paths:
        # mod_path = Path(src_path, *module.split('.')).with_suffix('.py')
        module = '.'.join(mod_path.parent.parts + (mod_path.stem,))
        if mod_path.exists():
            print(colorama.Fore.YELLOW, 'WARNING, file {} already exists, '.format(mod_path))
            mod_path = mod_path.with_suffix('.py.%s' %
                                            (dt.datetime.now().timestamp()))
            print('Writing in {} instead'.format(mod_path) + colorama.Style.RESET_ALL)
        mod_path.parent.mkdir(parents=True, exist_ok=True)
        print(colorama.Fore.GREEN + colorama.Style.BRIGHT +\
              'GENERATING CODE SKELETON {}\nFOR MODULE {}'.format(mod_path, module) +\
              colorama.Style.RESET_ALL)
        with open(mod_path, 'w') as f:
            print(model.generate_skeleton(module), file=f)


def run_model(params):
    """Run the model with the specified local parameters.

    Args:
        params: a dictionary with all parameters required to carry out the
          simulations

    Returns:
        The instance of MultiSimulation class which carried out the simulations

    See also:
        `emulsion.tools.simulation.MultiSimulation`_
    """
    count_path = Path(params.output_dir, 'counts.csv')
    if count_path.exists():
        count_path.unlink()
    log_path = Path(params.output_dir, 'log.txt')
    if log_path.exists():
        log_path.unlink()
    multi_simu = MultiSimulation(**params)
    # multi_simu.write_dot()
    start = time.perf_counter()
    multi_simu.run()
    end = time.perf_counter()
    print(colorama.Style.BRIGHT + 'Simulation finished in {:.2f} s'.format(end-start))
    print(colorama.Fore.GREEN + 'Outputs stored in {}'.format(count_path) + colorama.Style.RESET_ALL)
    return multi_simu


def produce_diagrams(params, view=False):
    """Use Graphviz to render state machines of the model. If *view* is
    set to True, opens the diagrams with system viewer.

    Args:
        params: a dictionary with all parameters required to carry out the
          simulations
        view: a boolean indicating whether or not diagrams have to be opened
          directly

    """
    model = params.model
    model.write_dot(params.output_dir)
    prefix = model.model_name
    for name, _ in model.state_machines.items():
        inpath = Path(params.output_dir, prefix + '_' + name + '.dot')
        outpath = Path(params.figure_dir,
                       prefix + '_' + name + '_machine.' + params.img_format)
        os.system("dot -T%s %s > %s" % (params.img_format, inpath, outpath))
        print(colorama.Fore.GREEN + 'Generated state machine diagram {}'.format(outpath) + colorama.Style.RESET_ALL)
        if view:
            webbrowser.open(outpath.absolute().as_uri())

def check_model(params, filemodel, view_meta=False, view_model=False):
    """Check the syntax of the model according to the grammar specified in
    EMULSION metamodel (``src/emulsion/resources/emulsion.tx``). If
    *show_meta* is True, produce a figure to represent the metamodel
    in *figure_dir* (requires GraphViz). If the syntax is correct and
    *show_model* is True, also produce a figure for the model
    structure.

    """
    source_path = Path(filemodel)
    metapath = get_metamodel()
    if not metapath.exists():
        print(colorama.Fore.RED + colorama.Style.BRIGHT +\
              'ERROR, metamodel file not found: {}'.format(metapath) +\
              colorama.Style.RESET_ALL)
        sys.exit(-1)
    meta = metamodel_from_file(metapath)
    if view_meta:
        meta_output = Path(params.figure_dir,
                           'meta_' + metapath.name).with_suffix('.dot')
        metamodel_export(meta, meta_output)
        figname = str(meta_output.with_suffix('.' + params.img_format))
        os.system("dot -T%s %s > %s" % (params.img_format,
                                        meta_output, figname))
        print(colorama.Fore.GREEN + 'Produced Metamodel figure: {}'.format(figname) +\
              colorama.Style.RESET_ALL)

    with open(source_path) as f:
        content = f.read()

    normalized = yaml.dump(yaml.load(content), default_flow_style=False)
    with open('tmp.yaml', 'w') as f:
        print(normalized, file=f)
    model_check = meta.model_from_str(normalized)
    if view_model:
        model_path = Path(params.figure_dir,
                          'model_%s' % (source_path.name, )).with_suffix('.dot')
        model_export(model_check, model_path)
        figname = str(model_path.with_suffix('.' + params.img_format))
        os.system("dot -T%s %s > %s" % (params.img_format, model_path, figname))
        print(colorama.Fore.GREEN + 'Produced Model figure {}'.format(figname) + colorama.Style.RESET_ALL)
    Path('tmp.yaml').unlink()
    print(colorama.Fore.GREEN + colorama.Style.BRIGHT +\
          'File {} complies with Emulsion syntax'.format(filemodel)+\
          colorama.Style.RESET_ALL)


def not_implemented(_):
    """Default behavior for unimplemented features.

    """
    print(colorama.Fore.RED + colorama.Style.BRIGHT +\
          'Feature not implemented in this model.' + colorama.Style.RESET_ALL)
    sys.exit(0)

def set_seed(params, seed=None, show=False):
    """Initialize the numpy's Random Number Generator, either with the
    specified seed, or with a seed calculated from current time and
    process ID.

    """
    if seed is None:
        params.seed = int(os.getpid() + time.time())
    else:
        params.seed = int(seed)
    np.random.seed(params.seed)
    if show:
        print(colorama.Style.BRIGHT + 'RANDOM SEED: {}'.format(params.seed) + colorama.Style.RESET_ALL)


def init_main_level(params):
    """Initialize the upper simulation level, in charge of making all
    sub-levels work properly.

    """
    if params.level not in params.model.levels:
        print(colorama.Fore.RED + colorama.Style.BRIGHT + \
              'ERROR, level {} not found'.format(params.level) + colorama.Style.RESET_ALL)
        sys.exit(-1)

    module_name = params.model.levels[params.level]['module']
    class_name = params.model.levels[params.level]['class_name']
    try:
        params.target_agent_class = load_class(module_name,
                                               class_name=class_name)[0]
    except AttributeError:
        print(colorama.Fore.RED + colorama.Style.BRIGHT + \
              'ERROR, agent class not found for level {}: {}.{}'.format(
                  params.level, module_name, class_name) + colorama.Style.RESET_ALL)
        sys.exit(-1)
    except ModuleNotFoundError:
        print(colorama.Fore.RED + colorama.Style.BRIGHT + \
              'ERROR, module not found for level {}: {}'.format(params.level, module_name) +\
              colorama.Style.RESET_ALL)
        sys.exit(-1)

def main(args=None):
    """Run EMuLSion's main program according to the command-line
    arguments.

    """
    colorama.init()
    if args is None:
        args = docopt(__doc__, version=get_version())

    if args['--license']:
        print(colorama.Fore.CYAN + get_license() + colorama.Style.RESET_ALL)
        sys.exit(0)

    if not Path(args['MODEL']).exists():
        print(colorama.Fore.RED + colorama.Style.BRIGHT +\
              'ERROR: file {} not found'.format(args['MODEL']) + colorama.Style.RESET_ALL)
        sys.exit(-1)

    params = StateVarDict()
    params.model = EmulsionModel(filename=args['MODEL'])
    params.nb_simu = int(args['--runs'])
    params.stochastic = not args['--deterministic']
    params.to_describe = args['PARAM']

    params.save_to_file = args['--save']
    params.load_from_file = args['--load']

    params.level = args['--level']
    if params.level is None:
        params.level = params.model.root_level
    if not args['--modifiable']:
        print(colorama.Style.DIM + 'Simulation level:{}'.format(params.level) + colorama.Style.NORMAL)

    params.silent = args['--silent']
    params.quiet = args['--quiet']
    params.log_params = args['--log-params']
    params.start_id = int(args['--start-id'])

    params.output_dir = Path(args['--output-dir'])
    params.input_dir = Path(args['--input-dir'])
    params.figure_dir = Path(args['--figure-dir'])
    params.img_format = args['--format']
    if not params.output_dir.exists():
        params.output_dir.mkdir(parents=True)
    if not params.figure_dir.exists():
        params.figure_dir.mkdir(parents=True)
    params.output_dir = str(params.output_dir)

    params.stock_agent = False
    params.keep_history = False

    if args['--param']:
        change_parameters(params, args['--param'])

    if args['--time']:
        params.steps = int(args['--time'])
    elif 'total_duration' in params.model.parameters:
        params.steps = int(np.ceil(params.model.get_value('total_duration') \
                                   / params.model.delta_t))
    else:
        params.steps = 100

    if args['--modifiable']:
        show_parameters(params, short=True)

    set_seed(params, seed=args['--seed'], show=args['--show-seed'])

    if args['--echo']:
        print(args)
        sys.exit(0)

    params.view_machines = False
    if args['--view-model']:
        produce_diagrams(params)
        params.view_machines = True

    if args['check']:
        check_model(params, args['MODEL'], args['--meta-tree'], args['--model-tree'])
    elif args['diagrams']:
        produce_diagrams(params, view=True)
        sys.exit(0)
    elif args['generate']:
        generate_model(params)
    elif args['run']:
        init_main_level(params)
        run_model(params)
        if args['--plot']:
            plot_outputs(params)
    elif args['show']:
        show_parameters(params)
    elif args['describe']:
        describe_parameters(params)
    elif args['plot']:
        plot_outputs(params)
    elif args['change']:
        not_implemented(params)
    elif args['sensitivity']:
        not_implemented(params)
    else:
        return params


################################################################
#                  _
#                 (_)
#  _ __ ___   __ _ _ _ __
# | '_ ` _ \ / _` | | '_ \
# | | | | | | (_| | | | | |
# |_| |_| |_|\__,_|_|_| |_|
#
################################################################

if __name__ == '__main__':
    main()



"""
.. module:: emulsion.agent.managers.functions

.. moduleauthor:: Sébastien Picault <sebastien.picault@inra.fr>

"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from   collections               import Counter

import numpy                     as np
import pandas                    as pd

from   emulsion.tools.misc       import select_random,\
    create_counter_getter, create_atoms_aggregator



from   emulsion.agent.managers.multi_process_manager  import  MultiProcessManager


#  _____ ____  __  __ _____
# |_   _|  _ \|  \/  |  __ \
#   | | | |_) | \  / | |__) | __ ___   ___ ___  ___ ___
#   | | |  _ <| |\/| |  ___/ '__/ _ \ / __/ _ \/ __/ __|
#  _| |_| |_) | |  | | |   | | | (_) | (_|  __/\__ \__ \
# |_____|____/|_|  |_|_|   |_|  \___/ \___\___||___/___/
#  __  __
# |  \/  |
# | \  / | __ _ _ __   __ _  __ _  ___ _ __
# | |\/| |/ _` | '_ \ / _` |/ _` |/ _ \ '__|
# | |  | | (_| | | | | (_| | (_| |  __/ |
# |_|  |_|\__,_|_| |_|\__,_|\__, |\___|_|
#                            __/ |
#                           |___/

class IBMProcessManager(MultiProcessManager):
    """An IBMProcessManager is a MultiProcessManager dedicated to the
    management of Individual-Based Models. This class is endowed with
    a `counters` attribute which is a dictionary {process -> counter of
    states in relation with the process}.

    """
    def __init__(self, **others):
        super().__init__(**others)
        self.statemachines = self.find_sublevel_statemachines()
        self.counters = {}
        self.autoremove_states = []
        self.new_agents = []
        self.counters = {}
        for machine in self.statemachines:
            self.counters[machine.machine_name] = Counter(
                atom.statevars[machine.machine_name].name
                for atom in self['MASTER']
            )
            self.autoremove_states += [(state.state_machine.machine_name, state)
                                       for state in machine.states
                                       if machine.get_property(state.name, 'autoremove')]
            for state in machine.states:
                if not state.autoremove:
                    self.create_count_properties_for_state(machine.machine_name,
                                                           state.name,
                                                           create_counter_getter,
                                                           create_atoms_aggregator)
        # print(self.autoremove_states)

        for atom in self['MASTER']:
            atom.set_statemachines(self.statemachines)
            atom.init_level_processes()

    def add_atoms(self, atom_set, init=False, **others):
        super().add_atoms(atom_set, init=init, **others)
        if not init:
            for atom in self['MASTER']:
                atom.set_statemachines(self.statemachines)
                atom.init_level_processes()

    def get_sublevels(self):
        """Return the list of sublevels contained in this level."""
        return self.model.levels[self.level]['contains']\
            if 'contains' in self.model.levels[self.level] else []

    def find_sublevel_statemachines(self):
        """Retrieve state machines used as processes by agents from the
        sub-level.

        """
        sublevels = self.get_sublevels()
        if not sublevels:
            return []
        return set([self.model.state_machines[process]
                    for sublevel in sublevels
                    for process in self.model.processes[sublevel]
                    if process in self.model.state_machines])

    def evolve(self, **others):
        """Make the agent evolve and update counts based on sub-level
        agents.

        """
        # prepair creation of new agents
        self.new_agents.clear()
        super().evolve(**others)
        ## handle autoremove states
        to_remove = []
        for machine_name, state in self.autoremove_states:
            to_remove += self.select_atoms(machine_name, value=state)
        self.remove_atoms(set(to_remove))
        self.add_new_population(None, self.new_agents)
        self.update_counts()


    def update_counts(self):
        """Update counters based on invdividual status."""
        for name, counter in self.counters.items():
            counter.clear()
            counter.update([atom.statevars[name].name
                            for atom in self['MASTER']])
        # print(self.counters)
    @property
    def counts(self):
        """Return a pandas DataFrame containing counts of each process if
        existing.

        """
        res = {state.name: self.counters[state_machine.machine_name][state.name]
               for state_machine in self.statemachines
               for state in state_machine.states}
        res.update({'step': self.statevars.step,
                    'level': self.level,
                    'agent_id': self.agid,
                    # 'population': self.population}
        })
        if self.level in self.model.outputs and\
           'extra_vars' in self.model.outputs[self.level]:
            res.update({name: self.get_model_value(name)\
                        if name in self.model.parameters\
                        else self.get_information(name)
                        for name in self.model.outputs[self.level]['extra_vars']})
        return pd.DataFrame(res, index=[0])

    def remove_randomly(self, proba=0, statevar=None):
        """Remove randomly chosen atoms from this ProcessManager. `proba` can
        be either a probability or a dictionary. In that case, the
        `statevar` parameter indicates the name of the state variable
        which drives the probabilities, and the keys must be valid
        values for this state variable. Selected atoms are removed and
        returned by the method.

        """
        if statevar is None:
            to_remove = select_random(self['MASTER'],
                                      np.random.binomial(len(self['MASTER']),
                                                         proba))
        else:
            to_remove = []
            for atom in self['MASTER']:
                val = atom.get_information(statevar)
                if val in proba:
                    if np.random.binomial(1, proba[val]):
                        to_remove.append(atom)
        self.remove_atoms(to_remove)
        # print(to_remove)
        return to_remove
